from __future__ import annotations

"""
A Model Mixin for SQL / Pydantic Models Designed for Improving Read Performance
by levearging `psqlpy`.

https://psqlpy-python.github.io/benchmarks.html
"""

import abc
import datetime
from lzl.types import BaseModel, Field, eproperty, rproperty
from lzl.logging import logger, null_logger
from pydantic.alias_generators import to_snake
import typing as t

if t.TYPE_CHECKING:
    from lzl.io import SerT
    from psqlpy import QueryResult
    from sqlalchemy.ext.asyncio import AsyncResult


SchemaT = t.TypeVar('SchemaT', bound = BaseModel)

class DBSchemaHandler(abc.ABC):
    """
    The Database Schema Handler

    It takes an existing BaseModel to handle Retrieval of Objects from the DB
    """

    table: t.Optional[str] = None
    schema_name: t.Optional[str] = None
    ref: t.Type[SchemaT] = None
    _extra: t.Dict[str, t.Any] = {}

    def __init__(
        self,
        model: t.Type[SchemaT],
        table: t.Optional[str] = None,
        schema_name: t.Optional[str] = None,
        debug_enabled: t.Optional[bool] = None,
        include_auto_types: t.Optional[bool] = False,
        autocast_to_schema: t.Optional[bool] = True,
        default_engine: t.Optional[t.Literal['psqlpy', 'sqlalchemy']] = 'psqlpy',
        **kwargs,
    ):
        """
        Initializes the PostgreSQL Schema Handler
        """
        self.ref = model
        self.table = table or to_snake(model.__name__)
        if schema_name: self.schema_name = schema_name
        self.logger = logger
        self.autologger = logger if debug_enabled is None else null_logger
        self.include_auto_types = include_auto_types
        self.autocast_to_schema = autocast_to_schema
        self.default_engine = default_engine
        self.postinit(**kwargs)

    def postinit(self, **kwargs):
        """
        Post Initialization
        """
        pass

    """
    Properties
    """

    @eproperty
    def table_name(self) -> str:
        """
        Returns the Table Name
        """
        return f'{self.schema_name}.{self.table}' if \
            self.schema_name else self.table

    @eproperty
    def columns(self) -> t.List[str]:
        """
        Returns the Columns
        """
        return [
            name for name, field in self.ref.model_fields.items()
            if not field.exclude and (
                not field.json_schema_extra or 
                not field.json_schema_extra.get('exclude_column', field.json_schema_extra.get('excluded_column'))
            )
        ]

    @eproperty
    def columns_as_string(self) -> str:
        """
        Returns the columns string
        """
        return ', '.join(self.columns)
    
    @eproperty
    def ser_columns(self) -> t.List[str]:
        """
        Returns the columns that are serialized
        """
        return [
            name for name, field in self.ref.model_fields.items()
            if field.annotation not in [
                str, t.Optional[str],
                int, t.Optional[int],
                float, t.Optional[float],
                bool, t.Optional[bool],
                datetime, t.Optional[datetime.datetime],
                list, t.Optional[t.List],
                dict, t.Optional[t.Dict],
            ]
            and name in self.columns
        ]

    @eproperty
    def _ser_columns_config(self) -> t.Dict[str, t.Dict[str, t.Union[str, int]]]:
        # sourcery skip: dict-comprehension, inline-immediately-returned-variable
        """
        Returns the Columns that are Serialized
        """
        # These are the columns that are serialized
        ser_columns = [name for name, field in self.ref.model_fields.items() if field.json_schema_extra and field.json_schema_extra.get('ser')]
        ser_dict = {}
        for name in ser_columns:
            ser_dict[name] = {
                'serializer': self.ref.model_fields[name].json_schema_extra.get('ser'),
                'compression': self.ref.model_fields[name].json_schema_extra.get('ser_comp'),
                'compression_level': self.ref.model_fields[name].json_schema_extra.get('ser_comp_level'),
            }
        # Use default serializers for the rest
        if self.include_auto_types:
            for name in self.ser_columns:
                if name in ser_dict: continue
                ser_dict[name] = {
                    'serializer': 'json',
                }
        return ser_dict

    @eproperty
    def _column_serializers(self) -> t.Dict[str, 'SerT']:
        """
        Returns the Serializers that are used for the schema
        """
        from lzl.io.ser import get_serializer
        return {
            name: get_serializer(**ser_config) for name, ser_config in self._ser_columns_config.items()
        }

    @eproperty
    def custom_decoders(self) -> t.Dict[str, t.Callable[..., t.Any]]:
        """
        Returns the Custom Decoders
        """
        return {
            name: ser.loads for name, ser in self._column_serializers.items()
        }


    """
    Helper Functions
    """

    @classmethod
    def normalize_ids_to_str(cls, ids: t.List[str]) -> str:
        """
        Normalizes the ids to a string
        """
        return ', '.join([f"'{id}'" for id in ids])

    def get_column_names(
        self,
        columns: t.Optional[t.List[str]] = None,
        **kwargs,
    ) -> t.List[str]:
        """
        Returns the Column Names
        """
        # This can be subclassed to handle additional filtering
        return self.columns if columns is None else columns
    
    def get_excluded_column_names(
        self,
        excluded_columns: t.Optional[t.List[str]] = None,
        **kwargs,
    ) -> t.List[str]:
        """
        Returns the Excluded Column Names
        """
        # This can be subclassed to handle additional filtering
        return [] if excluded_columns is None else excluded_columns


    def build_sql_filter(
        self,
        conditional: t.Optional[str] = 'AND',
        table_name: t.Optional[str] = None,
        include_table_name: t.Optional[bool] = None,
        **filters: t.Dict[str, t.Union[int, float, datetime.datetime, t.Dict, t.List, t.Any]]
    ) -> t.Dict[str, t.Union[t.List[str], str]]:
        """
        Returns 
        {
            "conditional": "AND",
            "statements": [
                "statement1",
                "statement2",
                "statement3",
            ]
        }
        which is properly formatted for the template
        """
        raise NotImplementedError
        return {
            "conditional": conditional,
            "statements": build_sql_filter_query_list(table_name = table_name, include_table_name = include_table_name, **filters),
        }

    def build_sql_filters(
        self,
        and_filters: t.Optional[t.Dict[str, t.Union[int, float, datetime.datetime, t.Dict, t.List, t.Any]]] = None,
        or_filters: t.Optional[t.Dict[str, t.Union[int, float, datetime.datetime, t.Dict, t.List, t.Any]]] = None,
        table_name: t.Optional[str] = None,
        include_table_name: t.Optional[bool] = None,
    ) -> t.List[t.Dict[str, t.Union[t.List[str], str]]]:
        """
        Creates the proper SQL filters
        [
            {
                "conditional": "AND",
                "statements": [
                    "statement1",
                    "statement2",
                    "statement3",
                ]
            },
            {
                "conditional": "OR",
                "statements": [
                    "statement4",
                    "statement5",
                    "statement6",
                ]
            }
        ]
        which is properly formatted for the template
        """
        table_name = table_name or self.table_name
        filters = []
        if and_filters: filters.append(self.build_sql_filter(conditional = "AND", table_name = table_name, include_table_name = include_table_name, **and_filters))
        if or_filters: filters.append(self.build_sql_filter(conditional = 'OR', table_name = table_name, include_table_name = include_table_name, **or_filters))
        return filters

    def build_filter_query(
        self,
        query: str,
        filters: t.Dict[str, t.Union[int, float, datetime.datetime, t.Dict, t.List, t.Any]],
    ) -> str:
        """
        Creates the filter query
        """
        query += '\nWHERE ' if 'WHERE' not in query else ' AND '
        filters = self.build_sql_filters(and_filters = filters)
        for fn, f in enumerate(filters):
            for sn, statement in enumerate(f['statements']):
                query += f'{statement} '
                if sn < len(f['statements']) - 1: query += f['conditional']
            if fn < len(filters) - 1: query += ' AND '
        return query
    

    def construct_select_query(
        self,
        columns: t.Optional[t.List[str]] = None,
        **kwargs,
    ) -> str:
        """
        Constructs the Select Query
        """
        # Can be subclassed to add additional columns
        columns = self.get_column_names(columns = columns, **kwargs)
        column_str = ', '.join(columns)
        return f'SELECT {column_str} FROM {self.table_name}'


    def build_select_query(
        self,
        ids: t.Optional[t.List[str]] = None,
        filters: t.Optional[t.Dict[str, t.Union[int, float, datetime.datetime, t.Dict, t.List, t.Any]]] = None,
        limit: t.Optional[int] = None,
        offset: t.Optional[int] = None,
        order_by: t.Optional[str] = None,
        order_direction: t.Optional[str] = 'ASC',
        columns: t.Optional[t.List[str]] = None,
        **kwargs,
    ) -> str:
        """
        Builds the get query
        """
        query = self.construct_select_query(columns = columns, **kwargs)
        if ids: 
            id_str = self.normalize_ids_to_str(ids)
            query += f'\nWHERE id IN ({id_str})'
        if filters: query = self.build_filter_query(query, filters)
        if order_by: query += f"\nORDER BY {order_by} {order_direction}"
        if limit: query += f'\nLIMIT {limit}'
        if offset: query += f'\nOFFSET {offset}'
        return query

    """
    The Iterators
    """

    def finalize_row(
        self,
        row: t.Tuple[t.Any, ...],
        new_row: t.Dict[str, t.Any],
        **kwargs,
    ) -> t.Dict[str, t.Any]:
        """
        Finalizes the row
        """
        return new_row

    def map_row_sa(
        self,
        row: t.Tuple[t.Any, ...],
        columns: t.Optional[t.List[str]] = None,
        excluded_columns: t.Optional[t.List[str]] = None,
        cast_to_schema: t.Optional[bool] = None,
        **kwargs,
    ) -> t.Union[SchemaT, t.Dict[str, t.Any]]:
        """
        Maps the row to the schema model (sqlalchemy)
        """
        columns = self.get_column_names(columns = columns, **kwargs)
        excluded_columns = self.get_excluded_column_names(excluded_columns = excluded_columns, **kwargs)
        new_row = {}
        for col, row in zip(columns, row):
            if col in excluded_columns: continue
            if col in self._column_serializers and row is not None:
                row = self._column_serializers[col].loads(row)
            new_row[col] = row
        new_row = self.finalize_row(row = row, new_row = new_row, **kwargs)
        if cast_to_schema is None: cast_to_schema = self.autocast_to_schema
        return self.ref.model_validate(new_row) if cast_to_schema else new_row

    async def aiterate_sa(
        self,
        stream: t.AsyncGenerator['AsyncResult', None],
        columns: t.Optional[t.List[str]] = None,
        excluded_columns: t.Optional[t.List[str]] = None,
        extra_data: t.Optional[t.Dict[str, t.List[t.Any]]] = None,
        cast_to_schema: t.Optional[bool] = None,
        **kwargs
    ) -> t.AsyncGenerator[t.Union[SchemaT, t.Dict[str, t.Any]], None]:
        """
        Iterates the results (sqlalchemy)
        """
        if extra_data: extra_data = {k: iter(v) for k, v in extra_data.items()}
        async for row in stream:
            iter_data = {}
            if extra_data:
                iter_data = {k: next(v) for k, v in extra_data.items()}
            yield self.map_row_sa(
                row, 
                columns = columns, 
                excluded_columns = excluded_columns, 
                cast_to_schema = cast_to_schema,
                **iter_data,
                **kwargs
            )
    

    def map_row_pgp(
        self,
        row: t.Dict[str, t.Any],
        columns: t.Optional[t.List[str]] = None,
        excluded_columns: t.Optional[t.List[str]] = None,
        cast_to_schema: t.Optional[bool] = None,
        **kwargs,
    ) -> t.Union[SchemaT, t.Dict[str, t.Any]]:
        """
        Maps the row to the schema model (psqlpy)
        """
        columns = self.get_column_names(columns = columns, **kwargs)
        excluded_columns = self.get_excluded_column_names(excluded_columns = excluded_columns, **kwargs)
        new_row = {
            k: v for k, v in row.items() if k not in excluded_columns and
            k in columns
        }
        new_row = self.finalize_row(row = row, new_row = new_row, **kwargs)
        if cast_to_schema is None: cast_to_schema = self.autocast_to_schema
        return self.ref.model_validate(new_row) if cast_to_schema else new_row
    
    async def aiterate_pgp(
        self,
        stream: t.AsyncGenerator['QueryResult', None],
        columns: t.Optional[t.List[str]] = None,
        excluded_columns: t.Optional[t.List[str]] = None,
        extra_data: t.Optional[t.Dict[str, t.List[t.Any]]] = None,
        cast_to_schema: t.Optional[bool] = None,
        **kwargs
    ) -> t.AsyncGenerator[t.Union[SchemaT, t.Dict[str, t.Any]], None]:
        """
        Iterates the results (psqlpy)
        """
        if extra_data: extra_data = {k: iter(v) for k, v in extra_data.items()}
        async for result in stream:
            rows: t.List[t.Dict[str, t.Any]] = result.result(
                custom_decoders = self.custom_decoders,
            )
            for row in rows:
                iter_data = {k: next(v) for k, v in extra_data.items()} if extra_data else {}
                yield self.map_row_pgp(
                    row, 
                    columns = columns, 
                    excluded_columns = excluded_columns, 
                    cast_to_schema = cast_to_schema,
                    **iter_data,
                    **kwargs
                )



    def aiterate(
        self,
        stream: t.AsyncGenerator[t.Union['QueryResult', 'AsyncResult'], None],
        columns: t.Optional[t.List[str]] = None,
        excluded_columns: t.Optional[t.List[str]] = None,
        extra_data: t.Optional[t.Dict[str, t.List[t.Any]]] = None,
        cast_to_schema: t.Optional[bool] = None,
        engine: t.Optional[t.Literal['psqlpy', 'sqlalchemy']] = None,
        **kwargs
    ) -> t.AsyncGenerator[t.Union[SchemaT, t.Dict[str, t.Any]], None]:
        """
        Iterates the results and maps it to the proper 
        """
        if engine is None: engine = self.default_engine
        if engine == 'psqlpy':
            return self.aiterate_pgp(
                stream = stream,
                columns = columns,
                excluded_columns = excluded_columns,
                extra_data = extra_data,
                cast_to_schema = cast_to_schema,
                **kwargs
            )
        return self.aiterate_sa(
            stream = stream,
            columns = columns,
            excluded_columns = excluded_columns,
            extra_data = extra_data,
            cast_to_schema = cast_to_schema,
            **kwargs
        )



from __future__ import annotations

"""Pydantic model capturing :class:`Client` configuration state."""

import typing as t

import httpx

from lzl.types import BaseModel, ConfigDict

from . import typed as ht


class ClientParams(BaseModel):
    """Lightweight serialisable copy of the arguments passed to :class:`Client`."""
    auth: t.Optional[ht.AuthTypes] = None
    params: t.Optional[ht.QueryParamTypes] = None
    headers: t.Optional[ht.HeaderTypes] = None
    cookies: t.Optional[ht.CookieTypes] = None
    verify: t.Optional[ht.VerifyTypes] = None
    cert: t.Optional[ht.CertTypes] = None
    http1: t.Optional[bool] = None
    http2: t.Optional[bool] = None
    proxies: t.Optional[ht.ProxiesTypes] = None

    mounts: t.Optional[t.Mapping[str, httpx._client.BaseTransport]] = None
    async_mounts: t.Optional[t.Mapping[str, httpx._client.AsyncBaseTransport]] = None

    timeout: t.Optional[
        t.Union[t.Optional[float], t.Tuple[t.Optional[float], t.Optional[float], t.Optional[float], t.Optional[float]],
        httpx._client.Timeout,
    ]] = httpx._client.DEFAULT_TIMEOUT_CONFIG
    follow_redirects: t.Optional[bool] = None
    retries: t.Optional[int] = None
    limits: t.Optional[httpx._client.Limits] = httpx._client.DEFAULT_LIMITS
    max_redirects: int = httpx._client.DEFAULT_MAX_REDIRECTS
    event_hooks: t.Optional[t.Mapping[str, t.List[t.Callable]]] = None
    async_event_hooks: t.Optional[t.Mapping[str, t.List[t.Callable]]] = None
    base_url: t.Optional[t.Union[str, httpx._client.URL]] = ""
    transport: t.Optional[httpx._client.BaseTransport] = None
    async_transport: t.Optional[httpx._client.AsyncBaseTransport] = None

    app: t.Optional[t.Callable] = None
    trust_env: t.Optional[bool] = None
    default_encoding: str = "utf-8"
    kwargs: t.Any = None
    debug: t.Optional[bool] = None
    soup_enabled: t.Optional[bool] = None

    model_config = ConfigDict(extra = 'allow', arbitrary_types_allowed = True)

    @property
    def sync_kwargs(self) -> t.Dict[str, t.Any]:
        """Return keyword arguments safe for :class:`httpx.Client`."""
        data = self.model_dump(
            exclude_none = True, 
            exclude = {'async_transport', 'async_mounts', 'async_event_hooks', 'soup_enabled', 'debug', 'retries', 'proxies'}
        )
        if self.proxies: data['proxy'] = self.proxies
        kwargs = data.pop('kwargs', None)
        if kwargs: data.update(kwargs)
        return data
    
    @property
    def async_kwargs(self) -> t.Dict[str, t.Any]:
        """Return keyword arguments safe for :class:`httpx.AsyncClient`."""
        data = self.model_dump(
            exclude_none = True, 
            exclude = {'soup_enabled', 'debug', 'retries', 'proxies'}
        )
        if self.proxies: data['proxy'] = self.proxies
        if data.get('async_transport'):
            data['transport'] = data.pop('async_transport', None)
        if data.get('async_mounts'):
            data['mounts'] = data.pop('async_mounts', None)
        if data.get('async_event_hooks'):
            data['event_hooks'] = data.pop('async_event_hooks', None)
        kwargs = data.pop('kwargs', None)
        if kwargs: data.update(kwargs)
        return data



from __future__ import annotations

"""
Base Token Types
"""

import time
from lzo.types import BaseModel, Field, eproperty, model_validator, Literal
from typing import Optional, List, Dict, Any, Union, Type, TYPE_CHECKING

TokenTypes = Literal['access_token', 'id_token', 'oauth2_token']

class AccessToken(BaseModel):
    iss: Optional[str] = None
    sub: Optional[str] = None
    aud: Optional[Union[List, str]] = None
    exp: Optional[int] = None
    iat: Optional[int] = None
    nbf: Optional[int] = None
    jti: Optional[str] = None
    typ: Optional[str] = None
    azp: Optional[str] = None
    scp: Optional[List[str]] = None
    scope: Optional[str] = None

    value: Optional[str] = None # The actual token value
    provider: Optional[str] = None

    def update(self, data: Dict[str, Any]):
        """
        Updates the payload
        """
        if not data.get('iat') or \
            data['iat'] < self.iat: return
        for key, value in data.items():
            if value and hasattr(self, key): setattr(self, key, value)
    
    @property
    def is_expired(self) -> bool:
        """
        Returns True if the Token is Expired
        """
        return self.exp < int(time.time()) if self.exp else False

    @classmethod
    def from_payload(cls, payload: Dict[str, Any], token: Optional[str] = None, provider: Optional[str] = None) -> 'AccessToken':
        """
        Creates the Access Token Payload from a payload
        """
        access_token = cls(**payload)
        if token is not None: access_token.value = token
        if provider is not None: access_token.provider = provider
        return access_token


    @model_validator(mode = 'after')
    def validate_access_token(self):
        """
        Validates the Access Token
        """
        if self.aud and isinstance(self.aud, str):
            self.aud = [self.aud]
        return self
    
    @property
    def identity(self) -> Optional[str]:
        """
        Returns the Identity
        """
        return self.sub

    @property
    def _token_type(self) -> TokenTypes:
        """
        Returns the Token Type
        """
        return 'access_token'

class IDToken(AccessToken):
    hd: Optional[str] = None
    email: Optional[str] = None
    email_verified: Optional[bool] = None
    at_hash: Optional[str] = None
    name: Optional[str] = None
    picture: Optional[str] = None
    given_name: Optional[str] = None
    family_name: Optional[str] = None


    @property
    def _token_type(self) -> TokenTypes:
        """
        Returns the Token Type
        """
        return 'id_token'

class OAuth2Token(BaseModel):
    
    expires_in: Optional[int] = None
    refresh_token: Optional[str] = None
    scope: Optional[str] = None
    token_type: Optional[str] = None

    access_token: Optional[Union[AccessToken, str]] = None
    id_token: Optional[Union[IDToken, str]] = None
    issued_at: Optional[int] = None
    expiration_ts: Optional[float] = None
    provider: Optional[str] = None
    # value: Optional[str] = None # The actual token value

    @model_validator(mode = 'after')
    def validate_oauth_token(self):
        """
        Sets the Expires At
        """
        if self.expiration_ts is None:
            if self.expires_in is not None:
                self.expiration_ts = time.time() + (self.expires_in - 10)
            elif self.issued_at is not None:
                self.expiration_ts = self.issued_at + (self.expires_in - 10)
        elif self.expires_in is not None:
            self.expires_in = int(self.expiration_ts - time.time())
        return self
    

    @property
    def is_expired(self) -> bool:
        """
        Returns True if the Token is Expired
        """
        return self.expiration_ts < time.time() if self.expiration_ts else False
    
    @eproperty
    def id_token_cls(self) -> Type[IDToken]:
        """
        Returns the ID Token Class
        """
        return IDToken
    
    @eproperty
    def access_token_cls(self) -> Type[AccessToken]:
        """
        Returns the Access Token Class
        """
        return AccessToken


    def set_access_token(
        self, 
        data: Dict[str, Any],
        token: Optional[str] = None,
    ) -> None:
        """
        Sets the Access Token
        """
        access_token = self.access_token_cls(**data)
        if token is not None: access_token.value = token
        elif isinstance(self.access_token, str):
            access_token.value = self.access_token
        self.access_token = access_token

    def set_id_token(
        self, 
        data: Dict[str, Any],
        token: Optional[str] = None,
    ) -> None:
        """
        Sets the ID Token
        """
        id_token = self.id_token_cls(**data)
        if token is not None: id_token.value = token
        elif isinstance(self.id_token, str):
            id_token.value = self.id_token
        self.id_token = id_token
    
    @property
    def identity(self) -> Optional[str]:
        """
        Returns the Identity
        """
        if 'identity' not in self._extra:
            if self.id_token and not isinstance(self.id_token, str) and self.id_token.sub:
                self._extra['identity'] = self.id_token.sub
            elif self.access_token and not isinstance(self.access_token, str) and self.access_token.sub:
                self._extra['identity'] = self.access_token.sub
        return self._extra.get('identity')

    @property
    def email(self) -> Optional[str]:
        """
        Returns the Email
        """
        if 'email' not in self._extra:
            if self.id_token and not isinstance(self.id_token, str) and self.id_token.email:
                self._extra['email'] = self.id_token.email
            elif self.access_token and not isinstance(self.access_token, str) and self.access_token.get('email'):
                self._extra['email'] = self.access_token.get('email')
        return self._extra.get('email')
    
    @eproperty
    def email_domain(self) -> Optional[str]:
        """
        Returns the Email Domain
        """
        if 'email_domain' not in self._extra:
            if self.id_token and not isinstance(self.id_token, str) and self.id_token.hd:
                self._extra['email_domain'] = self.id_token.hd
            elif self.access_token and not isinstance(self.access_token, str) and self.access_token.get('hd'):
                self._extra['email_domain'] = self.access_token.get('hd')
        return self._extra.get('email_domain')

    @property
    def roles(self) -> Optional[List[str]]:
        """
        Returns the Roles
        """
        if 'roles' not in self._extra:
            if self.id_token and not isinstance(self.id_token, str) and self.id_token.get('roles'):
                self._extra['roles'] = self.id_token.get('roles')
            elif self.access_token and not isinstance(self.access_token, str) and self.access_token.get('roles'):
                self._extra['roles'] = self.access_token.get('roles')
            
            if self._extra.get('roles') and isinstance(self._extra['roles'], list) and \
                isinstance(self._extra['roles'][0], dict):
                self._extra['roles'] = [role['name'] for role in self._extra['roles']]
        return self._extra.get('roles')

    @property
    def role(self) -> Optional[str]:
        """
        Returns the Role
        """
        if 'role' not in self._extra:
            if self.id_token and not isinstance(self.id_token, str) and self.id_token.get('role'):
                self._extra['role'] = self.id_token.get('role')
            elif self.access_token and not isinstance(self.access_token, str) and self.access_token.get('role'):
                self._extra['role'] = self.access_token.get('role')
            elif self.roles:
                self._extra['role'] = self.roles[0]
        return self._extra.get('role')
    
    @property
    def scopes(self) -> Optional[List[str]]:
        """
        Returns the Scopes
        """
        if not self.scope: return []
        return ' '.split(self.scope) if self.scope.strip() else []
    
    @eproperty
    def value(self) -> Optional[str]:
        """
        Returns the Value
        """
        if self.access_token and not isinstance(self.access_token, str) \
            and self.access_token.value:
            return self.access_token.value
        if self.id_token and not isinstance(self.id_token, str) \
            and self.id_token.value:
            return self.id_token.value
        return self.access_token
    
    @property
    def display_name(self) -> Optional[str]:
        """
        Returns the Display Name
        """
        if 'display_name' not in self._extra:
            if self.id_token and self.id_token.get('name'):
                self._extra['display_name'] = self.id_token.get('name')
            elif self.access_token and self.access_token.get('name'):
                self._extra['display_name'] = self.access_token.get('name')
        return self._extra.get('display_name')


    def merge(self, other: Dict[str, Any]):
        """
        Merges from a token payload
        """
        # This is likely a id_token
        if other.get('email'): 
            # print('Merging ID Token')
            self.id_token.update(other)
        elif other.get('sub'): 
            # print('Merging Access Token')
            self.access_token.update(other)
        return self

    def set_provider(self, provider: str):
        """
        Sets the provider
        """
        self.provider = provider
        if self.access_token and not isinstance(self.access_token, str): self.access_token.provider = provider
        if self.id_token and not isinstance(self.id_token, str): self.id_token.provider = provider

    @property
    def _token_type(self) -> TokenTypes:
        """
        Returns the Token Type
        """
        return 'oauth2_token'
    

    @classmethod
    def from_payload(cls, payload: Dict[str, Any], token: Optional[str] = None, provider: Optional[str] = None) -> 'AccessToken':
        """
        Creates the Oauth2 Token Payload from a payload
        """
        oauth2_token = cls(**payload)
        if provider is not None: oauth2_token.set_provider(provider)
        return oauth2_token

    def _refresh_state(self):
        """
        Helper method to return the access token and id token to proper 
        classes
        """
        if self.access_token and isinstance(self.access_token, dict):
            self.access_token = self.access_token_cls.from_payload(self.access_token, provider = self.provider)
        if self.id_token and isinstance(self.id_token, dict):
            self.id_token = self.id_token_cls.from_payload(self.id_token, provider = self.provider)


from __future__ import annotations

"""
Google OAuth2 Token
"""

from ...types.token import (
    AccessToken as BaseAccessToken,
    IDToken as BaseIDToken,
    OAuth2Token as BaseOAuth2Token,
    eproperty,
    Field,
    model_validator,
)

from typing import Optional, List, Dict, Any, Union, Type, TYPE_CHECKING


class AccessToken(BaseAccessToken):
    """
    Google Access Token
    """
    email: Optional[str] = None
    email_verified: Optional[bool] = None
    provider: Optional[str] = 'google'


class IDToken(BaseIDToken):
    """
    Google ID Token
    """
    provider: Optional[str] = 'google'


class OAuth2Token(BaseOAuth2Token):
    """
    Google OAuth2 Token
    """
    access_token: Optional[Union[AccessToken, str]] = None
    id_token: Optional[Union[IDToken, str]] = None

    @eproperty
    def id_token_cls(self) -> Type[IDToken]:
        """
        Returns the ID Token Class
        """
        return IDToken
    
    @eproperty
    def access_token_cls(self) -> Type[AccessToken]:
        """
        Returns the Access Token Class
        """
        return AccessToken


from typing import List, Dict


class COLOR:
    """
    Color Constants
    """
    RED = '\033[91m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    BLUE = '\033[94m'
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    WHITE = '\033[97m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'



from __future__ import annotations

import datetime
import contextlib
import functools
from lzl.api import aiohttpx
from lzo.types import (
    BaseModel,
    eproperty,
    Field,
)
from lzo.types.base import (
    get_pydantic_field_names,
    PYDANTIC_VERSION,
)

from pydantic import RootModel
from typing import Dict, Optional, Any, List, Type, Union, Generator, AsyncGenerator, cast
from lzl.api.openai.types.errors import error_handler
from lzl.api.openai.types.base import BaseResource, Usage, FileObject
from lzl.api.openai.utils import logger

__all__ = [
    'BaseResponse',
    'BaseBatchResponseItem',
    'BatchBaseResponse',
]


"""
Response Class
"""



if PYDANTIC_VERSION == 2:
    from pydantic import PrivateAttr, field_validator as _validator
    prevalidator = functools.partial(_validator, mode = 'before')
else:
    from pydantic import validator as _validator
    prevalidator = functools.partial(_validator, pre = True)


class BaseResponse(BaseResource):
    
    """
    Base Object class for responses to
    inherit from
    """

    id: Optional[str] = None
    object: Optional[str] = None
    created: Optional[datetime.datetime] = None
    updated: Optional[datetime.datetime] = None
    
    choices: Optional[List[Type[BaseResource]]] = None
    data: Optional[Union[Type[BaseResource], List[BaseResource]]] = None # Used for single retrieval if not in list.
    events: Optional[List[BaseResource]] = None

    data_model: Optional[Type[BaseResource]] = None
    choice_model: Optional[Type[BaseResource]] = None
    event_model: Optional[Type[BaseResource]] = None

    # Other Metadata
    usage: Optional[Usage] = None
    model: Optional[str] = None

    input_object: Optional[BaseResource] = Field(exclude = True)
    response: Optional[aiohttpx.Response] = Field(exclude = True)
    response_data: Optional[Union[Dict, List, Any]] = Field(default = None, exclude = True)

    if PYDANTIC_VERSION == 2:
        _has_metadata: Optional[bool] = PrivateAttr(default = False)
        _stream_consumed: Optional[bool] = PrivateAttr(default = False)
        _stream_chunks: Optional[List[Any]] = PrivateAttr(default = None)
    else:
        _has_metadata: Optional[bool] = Field(default = False, exclude = True)
        _stream_consumed: Optional[bool] = Field(default = False, exclude = True)
        _stream_chunks: Optional[List[Any]] = Field(default = None, exclude = True)


    @property
    def has_choices(self):
        """
        Returns whether the response has choices
        """
        return self.choice_model is not None
    
    @property
    def has_data(self):
        """
        Returns whether the response has data
        """
        return self.data_model is not None
    
    @property
    def has_events(self):
        """
        Returns whether the response has events
        """
        return self.event_model is not None
    
    @eproperty
    def resource_model(self) -> Type[BaseResource]:
        """
        Returns the appropriate resource model
        """
        if self.has_choices:
            return self.choice_model
        if self.has_data:
            return self.data_model
        return self.event_model if self.has_events else None

    @eproperty
    def excluded_params(self) -> List[str]:
        return [
            "data_model", "choice_model", "event_model", 
            "excluded_params", "resource_model", "input_model", "metadata_fields",
            "_stream_consumed", "_stream_chunks", "_has_metadata", 
            "response", "response_data", "input_object",
        ]
    
    def dict(self, *args, exclude: Any = None, **kwargs):
        """
        Returns the dict representation of the response
        """
        if exclude is None: exclude = set()
        exclude = set(exclude) | set(self.excluded_params)
        return super().dict(*args, exclude = exclude, **kwargs)


    """
    Metadata Properties
    """
    @eproperty
    def metadata_fields(self) -> List[str]:
        """
        Returns the metadata fields
        """
        return [
            name for name in get_pydantic_field_names(self) if \
                name not in [
                    "response", "response_data", "input_object",
                    "data_model",  "choice_model", "event_model",
                    "data", "choices", "events"
                ]
        ]


    @property
    def since_seconds(self) -> int:
        """
        Returns the number of seconds since the response was created
        """
        return (datetime.datetime.now(datetime.timezone.utc) - self.created).total_seconds() \
            if self.created else -1
    
    @property
    def headers(self) -> Dict[str, str]:
        """
        Returns the response headers
        """
        return self.response.headers if self.response else {}
    
    @property
    def response_json(self) -> Dict:
        """
        Returns the response json
        """
        with contextlib.suppress(Exception):
            return self.response.json() if self.response else {}
        return BaseResource.handle_json(self.response.content) if self.response else {}

    @property
    def has_stream(self):
        """
        Returns whether the response is a stream
        """
        return "text/event-stream" in self.headers.get("content-type", "")
    
    @property
    def request_id(self) -> Optional[str]:
        """
        Returns the request id
        """
        return self.headers.get("request-id", self.headers.get("x-request-id"))
    
    @property
    def organization(self) -> str:
        """
        Returns the organization
        """
        return self.headers.get("openai-organization")
    
    @property
    def is_azure(self) -> bool:
        """
        Returns whether the response is from azure
        """
        return bool(self.headers.get("azureml-model-session", self.headers.get("azureml-model-group")))

    @property
    def response_ms(self) -> Optional[int]:
        """
        Returns the response time in ms
        """
        h = self.headers.get("openai-processing-ms")
        return None if h is None else round(float(h))
    
    @property
    def response_timestamp(self) -> Optional[datetime.datetime]:
        """
        Returns the response timestamp
        """
        dt = self.headers.get("date")
        return None if dt is None else datetime.datetime.strptime(dt, '%a, %d %b %Y %H:%M:%S GMT')

    @property
    def openai_version(self) -> Optional[str]:
        """
        Returns the openai version
        """
        return self.headers.get("openai-version")
    
    @property
    def openai_model(self) -> Optional[str]:
        """
        Returns the openai model
        """
        return self.headers.get("openai-model")
    
    @property
    def model_name(self) -> Optional[str]:
        """
        Returns the model name
        """
        return self.model.split('-')[1]
    
    @property
    def consumption(self) -> Optional[float]:
        """
        Returns the consumption cost
        """
        return None
    
    @eproperty
    def consumption_cost(self) -> str:
        """
        Returns the consumption cost for the completions
        """
        return f'${self.consumption:.5f}' if self.consumption else '$0.00'
    
    """
    Object Data Properties
    """

    @property
    def resource_data(self) -> List[Union[Type[BaseResource], Dict, Any]]:
        """
        Returns the appropriate data object
        """
        if self.has_data: return self.data
        if self.has_choices: return self.choices
        return self.events if self.has_events else []
    

    def __len__(self):
        return len(self.resource_data)
    
    def __iter__(self):
        return iter(self.resource_data)
    
    def __getitem__(self, key: Union[str, int]) -> Union[Type[BaseModel], Dict, Any]:
        """
        Returns the appropriate data object from the response to support more
        native access to the response data
        """
        if isinstance(key, str):
            if not self.response_data:
                self.response_data = self.response.json()
            return self.response_data.get(key)
        return self.resource_data[key] if self.resource_data else None
    
    """
    Response Handling Methods
    """

    def handle_metadata(
        self,
        item: Dict,
        **kwargs
    ):
        """
        Sets the appropriate metadata fields
        for the response resource object
        """
        for field in self.metadata_fields:
            if field in item:
                if field == 'usage':
                    self.usage = Usage.model_validate(item[field])
                    continue
                
                if 'created' in field or 'updated' in field:
                    # Handle created_at and updated_at fields as well
                    key = field.split('_', 1)[0]
                    setattr(self, key, datetime.datetime.fromtimestamp(item[field], datetime.timezone.utc))
                    continue

                if 'files' in field:
                    setattr(self, field, FileObject.create_many(item[field]))
                    continue

                setattr(self, field, cast(type(getattr(self, field)), item[field])) # type: ignore

        self._has_metadata = True

    def handle_data_item(
        self,
        item: Union[Dict, Any],
        **kwargs
    ):
        """
        Handle a single `data` item
        """
        if isinstance(item, dict) and item.get('data'):
            if self.data is None: self.data = []
            data = self.data_model.create_many(item['data'])
            self.data.extend(data)
        
        elif not self.data:
            self.data_model.model_validate(item) if self.data_model else item

    
    def handle_choice_item(
        self,
        item: Union[Dict, Any],
        **kwargs
    ):
        """
        Handle a single `choice` item
        """
        if self.choices is None: self.choices = []
        
        if item.get('choices'):
            choices = self.choice_model.create_many(item['choices'])
            self.choices.extend(choices)
        elif self.has_stream:
            self.choices.append(self.choice_model.model_validate(item))

        
    def handle_event_item(
        self,
        item: Union[Dict, Any],
        **kwargs
    ):
        """
        Handle a single `event` item
        """
        if item.get('events'):
            if self.events is None: self.events = []
            events = self.event_model.create_many(item['events'])
            self.events.extend(events)

    def handle_resource_item(
        self, 
        item: Union[Dict, Any],
        **kwargs
    ):
        """
        Handles a single resource item
        and routes it to the appropriate handler
        """
        if not self._has_metadata:
            self.handle_metadata(item, **kwargs)
        
        if self.has_choices and (item.get('choices') or self.has_stream):
            self.handle_choice_item(item, **kwargs)
        
        elif self.has_events and item.get('events'):
            self.handle_event_item(item, **kwargs)
        
        else:
            self.handle_data_item(item, **kwargs)
        
    def construct_resource(
        self, 
        parse_stream: Optional[bool] = True,
        **kwargs
    ):
        """
        Constructs the appropriate resource object
        from the response
        """
        if not self.has_stream: return self.handle_resource_item(item = self.response_json, **kwargs)
        if not parse_stream: return
        for item in self.handle_stream(
            response = self.response
        ):
            if "error" in item:
                raise error_handler(
                    response = self.response,
                    data = item,
                )
            
            self.handle_resource_item(item = item, **kwargs)
        self._stream_consumed = True


    async def aconstruct_resource(
        self, 
        parse_stream: Optional[bool] = True,
        **kwargs
    ):
        """
        Constructs the appropriate resource object
        from the response
        """
        if not self.has_stream: return self.handle_resource_item(item = self.response_json, **kwargs)
        if not parse_stream: return
        async for item in self.ahandle_stream(response=self.response):
            if "error" in item:
                raise error_handler(
                    response = self.response,
                    data = item,
                )
            
            self.handle_resource_item(item = item, **kwargs)
        self._stream_consumed = True
    
        
    @classmethod
    def prepare_response(
        cls, 
        response: aiohttpx.Response,
        input_object: 'BaseResource',
        parse_stream: Optional[bool] = True,
        **kwargs
    ) -> 'BaseResource':
        """
        Handles the response and returns the appropriate object
        """
        resource = cls(response = response, input_object = input_object)
        resource.construct_resource(parse_stream = parse_stream, **kwargs)
        return resource
    
    @classmethod
    async def aprepare_response(
        cls, 
        response: aiohttpx.Response,
        input_object: 'BaseResource',
        parse_stream: Optional[bool] = True,
        **kwargs
    ) -> 'BaseResource':
        """
        Handles the response and returns the appropriate object
        """
        resource = cls(response = response, input_object = input_object)
        await resource.aconstruct_resource(parse_stream = parse_stream, **kwargs)
        return resource


    def handle_stream_metadata(
        self,
        item: Dict[str, Any],
        **kwargs
    ):
        """
        Handles the stream metadata
        """
        if not self.id and item.get('id'):
            self.id = item['id']
        if not self.created and item.get('created'):
            self.created = datetime.datetime.fromtimestamp(item['created'], datetime.timezone.utc)
        if not self.usage:
            self.usage = Usage(completion_tokens=0)
    
    @eproperty
    def choices_results(self) -> List[Type[BaseResource]]:
        """
        Parses the choices first
        """
        if self.choices is None:
            self.construct_resource()
        return self.choices

    """
    New Handling of Stream Specific Responses
    """

    def parse_stream_item(self, item: Union[Dict, Any], **kwargs) -> Any:
        """
        Parses a single stream item
        """
        return item

    def stream(self, **kwargs) -> Generator[Dict, None, None]:
        """
        Handles the stream
        """
        if self._stream_consumed:
            yield from self._stream_chunks
            return
        self._stream_chunks = []
        for item in self.handle_stream(response = self.response, streaming = True):
            if "error" in item:
                raise error_handler(response = self.response, data = item)
            # self.handle_resource_item(item = item, **kwargs)
            stream_item = self.parse_stream_item(item = item, **kwargs)
            if stream_item is not None:
                self._stream_chunks.append(stream_item)
                yield stream_item
        self._stream_consumed = True
    
    async def astream(self, **kwargs) -> AsyncGenerator[Dict, None]:
        """
        Handles the stream
        """
        if self._stream_consumed:
            for stream_item in self._stream_chunks:
                yield stream_item
            return
        self._stream_chunks = []
        async for item in self.ahandle_stream(response = self.response, streaming = True):
            if "error" in item:
                raise error_handler(response = self.response, data = item)
            # self.handle_resource_item(item = item, **kwargs)
            stream_item = self.parse_stream_item(item = item, **kwargs)
            if stream_item is not None:
                self._stream_chunks.append(stream_item)
                yield stream_item
        self._stream_consumed = True
        

class BatchResponseData(BaseModel):
    """
    The response data for a batch request
    """
    status_code: Optional[int] = None
    request_id: Optional[str] = None
    body: Optional[Dict[str, Any]] = Field(default_factory = dict, description = "The body of the response")
    usage: Optional[Usage] = Field(default_factory = Usage, description = "The usage of the response")
    system_fingerprint: Optional[str] = None
    

class BaseBatchResponseItem(BaseResponse):
    """
    The response for a single item in a batch request
    """
    custom_id: Optional[str] = None
    input_object: Optional[BaseResource] = Field(None, exclude = True)
    response: Optional[BatchResponseData] = Field(default_factory = BatchResponseData, description = "The response for the item")
    error: Optional[Any] = Field(default = None, description = "The error for the item")


    """
    Modified Properties to support batch requests
    """
        
    @property
    def has_stream(self):
        """
        Returns whether the response is a stream
        """
        return False
    
    @property
    def headers(self) -> Optional[Dict[str, str]]:
        """
        Returns the response headers
        """
        return self._extra.get('headers', {})

    @headers.setter
    def headers(self, value: Dict[str, str]):
        """
        Sets the response headers
        """
        self._extra['headers'] = value
    
    @property
    def response_json(self) -> Dict:
        """
        Returns the response json
        """
        return self.response.body

    @property
    def request_id(self) -> Optional[str]:
        """
        Returns the request id
        """
        return self.response.request_id
    
    @property
    def organization(self) -> str:
        """
        Returns the organization
        """
        return self.headers.get("openai-organization")
    
    @property
    def is_azure(self) -> bool:
        """
        Returns whether the response is from azure
        """
        return bool(self.headers.get("azureml-model-session", self.headers.get("azureml-model-group")))

    @property
    def response_ms(self) -> Optional[int]:
        """
        Returns the response time in ms
        """
        h = self.headers.get("openai-processing-ms")
        return None if h is None else round(float(h))
    
    @property
    def response_timestamp(self) -> Optional[datetime.datetime]:
        """
        Returns the response timestamp
        """
        dt = self.headers.get("date")
        return None if dt is None else datetime.datetime.strptime(dt, '%a, %d %b %Y %H:%M:%S GMT')

    @property
    def openai_version(self) -> Optional[str]:
        """
        Returns the openai version
        """
        return self.headers.get("openai-version")
    
    @property
    def openai_model(self) -> Optional[str]:
        """
        Returns the openai model
        """
        return self.headers.get("openai-model", self.response.body.get('model', None))
    
    @property
    def model_name(self) -> Optional[str]:
        """
        Returns the model name
        """
        return self.model.split('-')[1]


    """
    Object Data Properties
    """

    @property
    def resource_data(self) -> List[Union[Type[BaseResource], Dict, Any]]:
        """
        Returns the appropriate data object
        """
        if self.has_data: return self.data
        if self.has_choices: return self.choices
        return self.events if self.has_events else []
    
    
    def __getitem__(self, key: Union[str, int]) -> Union[Type[BaseModel], Dict, Any]:
        """
        Returns the appropriate data object from the response to support more
        native access to the response data
        """
        if isinstance(key, str):
            if not self.response_data:
                self.response_data = self.response.body
            return self.response_data.get(key)
        return self.resource_data[key] if self.resource_data else None
    
    """
    Response Handling Methods
    """

    def construct_resource(
        self, 
        parse_stream: Optional[bool] = False,
        **kwargs
    ):
        """
        Constructs the appropriate resource object
        from the response
        """
        return self.handle_resource_item(item = self.response_json, **kwargs)


    async def aconstruct_resource(
        self, 
        parse_stream: Optional[bool] = False,
        **kwargs
    ):
        """
        Constructs the appropriate resource object
        from the response
        """
        return self.handle_resource_item(item = self.response_json, **kwargs)
    
        
    @classmethod
    def prepare_response(
        cls, 
        response_item: Dict[str, Any],
        input_object: 'BaseResource',
        parse_stream: Optional[bool] = False,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ) -> 'BaseBatchResponseItem':
        """
        Handles the response and returns the appropriate object
        """
        resource = cls.model_validate(response_item)
        resource.input_object = input_object
        if headers: resource.headers = headers
        resource.construct_resource(parse_stream = parse_stream, **kwargs)
        return resource
    
    @classmethod
    async def aprepare_response(
        cls, 
        response_item: Dict[str, Any],
        input_object: 'BaseResource',
        parse_stream: Optional[bool] = False,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ) -> 'BaseBatchResponseItem':
        """
        Handles the response and returns the appropriate object
        """
        resource = cls.model_validate(response_item)
        resource.input_object = input_object
        if headers: resource.headers = headers
        await resource.aconstruct_resource(parse_stream = parse_stream, **kwargs)
        return resource



class BatchBaseResponse(RootModel):
    """
    The base response for batch requests
    """
    root: List[BaseBatchResponseItem]

    def __len__(self):
        """
        Returns the length of the root
        """
        return len(self.root)


SUCCESSFUL_STATUSES = [
    'completed',
]
UNSUCESSFUL_STATUSES = [
    'failed',
    'expired',
    'cancelled',
]
IN_PROGRESS_STATUSES = [
    'validating',
    'in_progress',
    'finalizing',
    'cancelling',
]

TERMINAL_STATUSES = SUCCESSFUL_STATUSES + UNSUCESSFUL_STATUSES


class BatchStatusResponse(BaseModel):
    """
    The response for a single item in a batch request
    """
    id: str = Field(..., description = "The ID of the batch request")
    object: str = Field(..., description = "The type of the batch request")
    endpoint: str = Field(..., description = "The endpoint of the batch request")
    errors: Optional[Any] = Field(None, description = "The errors of the batch request")
    input_file_id: Optional[str] = Field(None, description = "The input file ID of the batch request")
    completion_window: Optional[str] = Field(None, description = "The completion window of the batch request")
    status: Optional[str] = Field(None, description = "The status of the batch request")
    output_file_id: Optional[str] = Field(None, description = "The output file ID of the batch request")
    error_file_id: Optional[str] = Field(None, description = "The error file ID of the batch request")
    created_at: Optional[datetime.datetime] = Field(None, description = "The created at of the batch request")
    in_progress_at: Optional[datetime.datetime] = Field(None, description = "The in progress at of the batch request")
    expires_at: Optional[datetime.datetime] = Field(None, description = "The expires at of the batch request")
    completed_at: Optional[datetime.datetime] = Field(None, description = "The completed at of the batch request")
    failed_at: Optional[datetime.datetime] = Field(None, description = "The failed at of the batch request")
    expired_at: Optional[datetime.datetime] = Field(None, description = "The expired at of the batch request")
    request_counts: Optional[Dict[str, int]] = Field(None, description = "The request counts of the batch request")
    metadata: Optional[Dict[str, Any]] = Field(None, description = "The metadata of the batch request")

    request_idx_mapping: Dict[str, int] = Field(default_factory = dict, description = "The request index mapping of the batch request")

    @prevalidator('created_at', 'in_progress_at', 'expires_at', 'completed_at', 'failed_at', 'expired_at')
    def validate_datetime(cls, v: Optional[Union[int, datetime.datetime]]) -> Optional[datetime.datetime]:
        """
        Validates the datetime
        """
        if not v: return None
        if not isinstance(v, datetime.datetime): v = datetime.datetime.fromtimestamp(v, datetime.timezone.utc)
        return v

    @property
    def is_completed(self) -> bool:
        """
        Returns whether the batch request is completed
        """
        return self.status in TERMINAL_STATUSES

    @property
    def is_in_progress(self) -> bool:
        """
        Returns whether the batch request is in progress
        """
        return self.status in IN_PROGRESS_STATUSES

    @property
    def is_successful(self) -> bool:
        """
        Returns whether the batch request is successful
        """
        return self.status in SUCCESSFUL_STATUSES

    @property
    def is_unsuccessful(self) -> bool:
        """
        Returns whether the batch request is unsuccessful
        """
        return self.status in UNSUCESSFUL_STATUSES


from __future__ import annotations

"""
OpenAI Functions Manager
"""

import functools
from abc import ABC
from lzl.proxied import ProxyObject
from lzo.utils import Timer
from lzl.types import eproperty
from lzl.api.openai.schemas.functions import (
    FunctionSchemaT,
    BaseFunction,
    FunctionT,
    FunctionResultT,

)
from typing import Optional, Any, Set, Dict, List, Union, Type, Tuple, Awaitable, Generator, AsyncGenerator, TypeVar, TYPE_CHECKING
if TYPE_CHECKING:
    from lzl.api.openai.configs import OpenAISettings
    from lzl.api.openai.clients import OpenAIManager as OpenAISessionManager
    from lzl.logging import Logger
    from lzl.io import PersistentDict



class FunctionManager(ABC):
    """
    The Functions Manager Class that handles registering and managing functions

    - Additionally supports caching through `kvdb`
    """

    name: Optional[str] = 'functions'

    def __init__(
        self,
        **kwargs,
    ):
        self._extra: Dict[str, Any] = {}
        # from lzl.api.openai.configs import settings
        # from lzl.api.openai.utils.logs import logger, null_logger
        # self.logger = logger
        # self.null_logger = null_logger
        # self.settings = settings
        # self.debug_enabled = self.settings.debug_enabled
        # self.cache_enabled = self.settings.function_cache_enabled

        self._api: Optional['OpenAISessionManager'] = None
        self._cache: Optional['PersistentDict'] = None
        self.functions: Dict[str, 'BaseFunction'] = {}
        self._kwargs = kwargs
        from lzo.utils.hashing import create_hash_from_args_and_kwargs
        self.create_hash = create_hash_from_args_and_kwargs

        # try:
        #     import xxhash
        #     self._hash_func = xxhash.xxh64
        # except ImportError:
        #     from hashlib import md5
        #     self._hash_func = md5
        
        try:
            import cloudpickle
            self._pickle = cloudpickle
        except ImportError:
            import pickle
            self._pickle = pickle

    @eproperty
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        from lzl.api.openai.utils.logs import logger
        return logger
            
    @eproperty
    def null_logger(self) -> 'Logger':
        """
        Returns the null logger
        """
        from lzl.api.openai.utils.logs import null_logger
        return null_logger

    @property
    def autologger(self) -> 'Logger':
        """
        Returns the logger
        """
        return self.logger if \
            (self.debug_enabled or self.settings.is_development_env) else self.null_logger

    @eproperty
    def settings(self) -> 'OpenAISettings':
        """
        Returns the settings
        """
        from lzl.api.openai.configs import settings
        return settings

    @eproperty
    def debug_enabled(self) -> bool:
        """
        Returns the debug enabled
        """
        return self.settings.debug_enabled

    @eproperty
    def cache_enabled(self) -> bool:
        """
        Returns the cache enabled
        """
        return self.settings.function_cache_enabled

    @property
    def api(self) -> 'OpenAISessionManager':
        """
        Returns the API
        """
        if self._api is None:
            from lzl.api.openai.clients import OpenAI
            # from async_openai.client import OpenAIManager
            self._api = OpenAI
        return self._api
    
    @property
    def cache(self) -> 'PersistentDict':
        """
        Gets the cache
        """
        if self._cache is None:
            serializer_kwargs = {
                'compression': self._kwargs.get('serialization_compression', None),
                'compression_level': self._kwargs.get('serialization_compression_level', None),
                'raise_errors': True,
            }
            kwargs = {
                'base_key': f'lzl.openai.functions.{self.api.settings.app_env.name}.{self.api.settings.proxy.proxy_app_name or "default"}',
                'expiration': self._kwargs.get('cache_expiration', 60 * 60 * 24 * 3),
                'serializer': self._kwargs.get('serialization', 'json'),
                'serializer_kwargs': serializer_kwargs,
            }
            try:
                import kvdb
                self._cache = kvdb.create_persistence(session_name = 'openai', **kwargs)
            except ImportError:
                from lzl.io.persistence import PersistentDict
                self._cache = PersistentDict(**kwargs)
        return self._cache

    def register_function(
        self,
        func: Union['BaseFunction', Type['BaseFunction'], str],
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        raise_error: Optional[bool] = False,
        initialize: Optional[bool] = True,
        return_initialized: Optional[bool] = False,
        **kwargs,
    ) -> Optional['BaseFunction']:
        """
        Registers the function
        """
        if isinstance(func, str):
            from lzl.load import lazy_import
            func = lazy_import(func)
        if isinstance(func, type) and initialize:
            func = func(**kwargs)
        name = name or func.name
        if not overwrite and name in self.functions:
            if raise_error: raise ValueError(f"Function {name} already exists")
            return self.functions[name] if return_initialized else None
        self.functions[name] = func
        self.logger.info(f"Registered Function: |g|{name}|e|", colored=True)
        return func if return_initialized else None

    async def acreate_hash(self, *args, **kwargs) -> str:
        """
        Creates a hash
        """
        return await self.api.pooler.asyncish(self.create_hash, *args, **kwargs)
    
    def _get_function(self, name: str) -> Optional['BaseFunction']:
        """
        Gets the function
        """
        func = self.functions.get(name)
        if func is None: return None
        if isinstance(func, type):
            func = func(**self._kwargs)
            self.functions[name] = func
        return func

    def get(self, name: Union[str, 'FunctionT', Type['FunctionT']]) -> 'FunctionT':
        """
        Gets the function
        """
        # print(name, type(name), isinstance(name, BaseFunction), issubclass(name, BaseFunction))
        # self.logger.info(f'Getting Function: {name}')
        if isinstance(name, str): return self._get_function(name)
        if isinstance(name, type): return self.register_function(name, initialize = True, return_initialized = True)
        if issubclass(name, BaseFunction):
            func = name(**self._kwargs)
            self.functions[name.name] = func
            return func    
        return name if isinstance(name, BaseFunction) else self._get_function(name)
        
    
    def parse_iterator_func(
        self,
        function: 'FunctionT',
        *args,
        with_index: Optional[bool] = False,
        **function_kwargs,
    ) -> Tuple[int, Set, Dict[str, Any]]:
        """
        Parses the iterator function kwargs
        """
        func_iter_arg = args[0]
        args = args[1:]
        idx = None
        if with_index: idx, item = func_iter_arg
        else: item = func_iter_arg
        _func_kwargs = function.get_function_kwargs()
        if isinstance(item, dict) and any(k in _func_kwargs for k in item):
            function_kwargs.update(item)
        else:
            # Get the missing function kwargs
            _added = False
            for k in _func_kwargs:
                if k not in function_kwargs:
                    function_kwargs[k] = item
                    # self.autologger.info(f"Added missing function kwarg: {k} = {item}", prefix = function.name, colored = True)
                    _added = True
                    break
            if not _added:
                # If not, then add the item as the first argument
                args = (item,) + args
        return idx, args, function_kwargs

    def execute(
        self,
        function: Union['FunctionT', str],
        *args,
        item_hashkey: Optional[str] = None,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]]:
        # sourcery skip: low-code-quality
        """
        Runs the function
        """
        overwrite = function_kwargs.pop('overwrite', None)
        overwrite = overwrite or overrides and 'functions' in overrides
        function = self.get(function)
        if overwrite and overrides and self.check_value_present(overrides, f'{function.name}.cachable'):
            cachable = False
        
        # Iterators
        is_iterator = function_kwargs.pop('_is_iterator', False)
        if is_iterator:
            idx, args, function_kwargs = self.parse_iterator_func(function, *args, with_index = with_index, **function_kwargs)
        
        if item_hashkey is None: item_hashkey = self.create_hash(**function_kwargs)
        key = f'{item_hashkey}.{function.name}'
        if function.has_diff_model_than_default: key += f'.{function.default_model_func}'
        if function.version: key += f'.{function.version}'

        t = Timer(format_short = 1)
        result = None
        cache_hit = False
        if self.cache_enabled and not overwrite:
            result: 'FunctionResultT' = self.cache.fetch(key)
            if result:
                if isinstance(result, dict): result = function.schema.model_validate(result)
                result.function_name = function.name
                cache_hit = True
        
        if result is None:
            result = function(*args, cachable = cachable, is_async = False, **function_kwargs)
            if self.cache_enabled and function.is_valid_response(result):
                self.cache.set(key, result)
        
        self.autologger.info(f"Function: {function.name} in {t.total_s} (Model: {result.function_model}, Client: {result.function_client_name}, Cache Hit: {cache_hit})", prefix = key, colored = True)
        if is_iterator and with_index:
            return idx, result if function.is_valid_response(result) else (idx, None)
        return result if function.is_valid_response(result) else None

    
    async def aexecute(
        self,
        function: Union['FunctionT', str],
        *args,
        item_hashkey: Optional[str] = None,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]]:
        # sourcery skip: low-code-quality
        """
        Runs the function
        """
        overwrite = function_kwargs.pop('overwrite', None)
        overwrite = overwrite or overrides and 'functions' in overrides
        function = self.get(function)

        if overwrite and overrides and self.check_value_present(overrides, f'{function.name}.cachable'):
            cachable = False
        
        # Iterators
        is_iterator = function_kwargs.pop('_is_iterator', False)
        if is_iterator:
            idx, args, function_kwargs = self.parse_iterator_func(function, *args, with_index = with_index, **function_kwargs)

        if item_hashkey is None: item_hashkey = await self.acreate_hash(*args, **function_kwargs)
        key = f'{item_hashkey}.{function.name}'
        if function.has_diff_model_than_default: key += f'.{function.default_model_func}'
        if function.version: key += f'.{function.version}'
        t = Timer(format_short = 1)
        result = None
        cache_hit = False
        if self.cache_enabled and not overwrite:
            result: 'FunctionResultT' = await self.cache.afetch(key)
            if result:
                if isinstance(result, dict): result = function.schema.model_validate(result)
                result.function_name = function.name
                cache_hit = True
        
        if result is None:
            result = await function(*args, cachable = cachable, is_async = True, **function_kwargs)
            if self.cache_enabled and function.is_valid_response(result):
                await self.cache.aset(key, result)
        
        self.autologger.info(f"Function: {function.name} in {t.total_s} (Model: {result.function_model}, Client: {result.function_client_name}, Cache Hit: {cache_hit})", prefix = key, colored = True)
        if is_iterator and with_index:
            return idx, result if function.is_valid_response(result) else (idx, None)
        return result if function.is_valid_response(result) else None

    
    
    @property
    def function_names(self) -> List[str]:
        """
        Returns the function names
        """
        return list(self.functions.keys())
    
    def __getitem__(self, name: str) -> Optional['FunctionT']:
        """
        Gets the function
        """
        return self.get(name)

    def __setitem__(self, name: str, value: Union[FunctionT, Type[FunctionT], str]):
        """
        Sets the function
        """
        return self.register_function(value, name = name)
    
    def append(self, value: Union[FunctionT, Type[FunctionT], str]):
        """
        Appends the function
        """
        return self.register_function(value)
    

    def check_value_present(
        self, items: List[str], *values: str,
    ) -> bool:
        """
        Checks if the value is present
        """
        if not values:
            return any(self.name in item for item in items)
        for value in values:
            key = f'{self.name}.{value}' if value else self.name
            if any((key in item or value in item) for item in items):
                return True
        return False
    
    def map(
        self,
        function: Union['FunctionT', str],
        iterable_kwargs: List[Union[Dict[str, Any], Any]],
        *args,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        return_ordered: Optional[bool] = True,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> List[Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]]]:
        """
        Maps the function to the iterable in parallel
        """
        partial = functools.partial(
            self.execute, 
            function, 
            # *args,
            cachable = cachable, 
            overrides = overrides, 
            _is_iterator = True,
            with_index = with_index,
            **function_kwargs
        )
        if with_index: iterable_kwargs = list(enumerate(iterable_kwargs))
        return self.api.pooler.map(partial, iterable_kwargs, *args, return_ordered = return_ordered)
    
    async def amap(
        self,
        function: Union['FunctionT', str],
        iterable_kwargs: List[Dict[str, Any]],
        *args,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        return_ordered: Optional[bool] = True,
        concurrency_limit: Optional[int] = None,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> List[Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]]]:
        """
        Maps the function to the iterable in parallel
        """
        partial = functools.partial(
            self.aexecute, 
            function, 
            # *args,
            cachable = cachable, 
            overrides = overrides, 
            _is_iterator = True,
            with_index = with_index,
            **function_kwargs
        )
        if with_index: iterable_kwargs = list(enumerate(iterable_kwargs))
        return await self.api.pooler.amap(partial, iterable_kwargs, *args, return_ordered = return_ordered, concurrency_limit = concurrency_limit)
    
    def iterate(
        self,
        function: Union['FunctionT', str],
        iterable_kwargs: List[Dict[str, Any]],
        *args,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        return_ordered: Optional[bool] = False,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> Generator[Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]], None, None]:
        """
        Maps the function to the iterable in parallel
        """
        partial = functools.partial(
            self.execute, 
            function, 
            # *args, 
            cachable = cachable, 
            overrides = overrides, 
            _is_iterator = True,
            with_index = with_index,
            **function_kwargs
        )
        if with_index: iterable_kwargs = list(enumerate(iterable_kwargs))
        return self.api.pooler.iterate(partial, iterable_kwargs, *args, return_ordered = return_ordered)

    def aiterate(
        self,
        function: Union['FunctionT', str],
        iterable_kwargs: List[Dict[str, Any]],
        *args,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        return_ordered: Optional[bool] = False,
        concurrency_limit: Optional[int] = None,
        with_index: Optional[bool] = False,
        **function_kwargs
    ) -> AsyncGenerator[Union[Optional['FunctionSchemaT'], Tuple[int, Optional['FunctionSchemaT']]], None]:
        """
        Maps the function to the iterable in parallel
        """
        partial = functools.partial(
            self.aexecute, 
            function, 
            # *args, 
            cachable = cachable, 
            overrides = overrides, 
            _is_iterator = True,
            with_index = with_index,
            **function_kwargs
        )
        if with_index: iterable_kwargs = list(enumerate(iterable_kwargs))
        return self.api.pooler.aiterate(partial, iterable_kwargs, *args, return_ordered = return_ordered, concurrency_limit = concurrency_limit)
    
    def __call__(
        self,
        function: Union['FunctionT', str],
        *args,
        item_hashkey: Optional[str] = None,
        cachable: Optional[bool] = True,
        overrides: Optional[List[str]] = None,
        is_async: Optional[bool] = True,
        **function_kwargs
    ) -> Union[Awaitable['FunctionSchemaT'], 'FunctionSchemaT']:
        """
        Runs the function
        """
        method = self.aexecute if is_async else self.execute
        return method(
            function = function,
            *args,
            item_hashkey = item_hashkey,
            cachable = cachable,
            overrides = overrides,
            **function_kwargs
        )
    



OpenAIFunctions: FunctionManager = ProxyObject(FunctionManager)


from __future__ import annotations

"""
Extended Types
"""

import contextlib
from sqlalchemy import types
from sqlalchemy.schema import Sequence
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.sql import operators
from sqlalchemy.dialects.postgresql import JSONB, ARRAY, JSON
from sqlalchemy.dialects.postgresql.base import ischema_names

from lzl import load
import typing as t

# if load.t.TYPE_CHECKING:
#     import fileio
# else:
#     fileio = load.LazyLoad("fileio", install_missing=True, install_options={'package': 'file-io'})

# if t.TYPE_CHECKING:
#     from fileio import FileLike

from lzl.io import File, FileLike

class FileField(types.TypeDecorator):

    impl = types.String
    cache_ok = True
    mysql_default_length = 255

    def coerce_compared_value(self, op, value):
        """
        Coerce the value to be compared.
        """
        return types.String(length = self.mysql_default_length) if op in (operators.like_op, operators.not_like_op) else self
    
    def process_bind_param(self, value: t.Optional[t.Union[str, 'FileLike']], dialect):
        """
        Return the value to be stored in the database.
        """
        if value is None: return None
        if isinstance(value, str): return value
        return value.as_posix() \
            if hasattr(value, 'as_posix') \
                else str(value)

    def process_result_value(self, value: t.Optional[t.Union[str, t.Any]], dialect):
        """
        Process the value retrieved from the database.
        """
        return File(value) if value is not None else None


class JsonString(types.TypeDecorator):
    """
    A String type that stores JSON data.
    """
    impl = types.String
    cache_ok = True
    hashable = True

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
        Initializes the JsonString t.Type
        """
        super().__init__(*args, **kwargs)
        from lzl.io.ser import get_serializer
        self.ser = get_serializer('json')
        self.ser.enforce_string_value = True

    def load_dialect_impl(self, dialect):
        """
        Load the dialect implementation.
        """
        return dialect.type_descriptor(self.impl)

    def process_bind_param(self, value, dialect) -> t.Optional[str]:
        """
        Process the value to be stored in the database.
        """
        if value is not None:
            value = self.ser.dumps(value)
        return value

    def process_result_value(self, value, dialect) ->  t.Optional[t.Union[t.Dict[str, t.Any], t.List[t.Dict[str, t.Any]]]]:
        """
        Process the value retrieved from the database.
        """
        if value is not None:
            value = self.ser.loads(value)
        return value


class SerializedBinary(types.TypeDecorator):
    """
    A LargeBinary type that stores serialized data.
    """
    impl = types.LargeBinary
    cache_ok = True
    hashable = True

    def __init__(
        self,
        *args,
        serializer: t.Optional[str] = 'json',
        compression: t.Optional[str] = 'zstd',
        compression_level: int = 3,
        validate_raw: t.Optional[bool] = True,
        **kwargs,
    ) -> None:
        """
        Initializes the SerializedBinary t.Type

        - Supports Serialization: json, pickle, msgpack
        - Supports Compression: zstd, lz4, zlib, and bz2
        """
        super().__init__(*args, **kwargs)
        from lzl.io.ser import get_serializer
        self.ser = get_serializer(serializer, compression = compression, compression_level = compression_level)
        self.validate_raw = validate_raw

    
    def load_dialect_impl(self, dialect):
        """
        Load the dialect implementation.
        """
        return dialect.type_descriptor(self.impl)

    def process_bind_param(self, value: t.Union[str, bytes, t.Any], dialect) -> t.Optional[bytes]:
        """
        Process the value to be stored in the database.
        """
        if value is not None:
            if self.validate_raw:
                if isinstance(value, bytes): return value
                if isinstance(value, str): return value.encode()
            value = self.ser.dumps(value)
        return value

    def process_result_value(self, value, dialect) ->  t.Optional[t.Union[t.Dict[str, t.Any], t.List[t.Dict[str, t.Any]]]]:
        """
        Process the value retrieved from the database.
        """
        if value is not None: value = self.ser.loads(value)
        return value



DateTimeUTC: types.DateTime = types.DateTime(timezone = True)

IntList: ARRAY = ARRAY(types.Integer, dimensions = 1)
BigIntList: ARRAY = ARRAY(types.BigInteger, dimensions = 1)
StringList: ARRAY = ARRAY(types.Text, dimensions = 1)
DateTimeUTCList: ARRAY = ARRAY(DateTimeUTC, dimensions = 1)
# AutoSequence: types.BigInteger = 

if load.t.TYPE_CHECKING:
    import sqlalchemy_json as sqla_json
    from sqlalchemy_json import NestedMutableJson, MutableDict

    JsonStringField: t.Type[MutableDict]
    JsonStringNestedField: t.Type[NestedMutableJson]

    JSONBField: t.Type[MutableDict] 
    JSONBNestedField: t.Type[NestedMutableJson]

else:
    sqla_json = load.lazy_load("sqlalchemy_json", install_missing=True)

    JsonStringField: t.Type[sqla_json.MutableDict] = sqla_json.mutable_json_type(dbtype = JsonString, nested=True)
    JsonStringNestedField: t.Type[sqla_json.NestedMutableJson] = sqla_json.mutable_json_type(dbtype = JsonString, nested = True) # type: ignore


    JSONBField: t.Type[sqla_json.MutableDict] = sqla_json.mutable_json_type(dbtype = JSONB)
    JSONBNestedField: t.Type[sqla_json.NestedMutableJson] = sqla_json.mutable_json_type(dbtype = JSONB, nested = True) # type: ignore


# with contextlib.suppress(ImportError):
#     from sqlalchemy_json import NestedMutableContainer, MutableContainer, TrackedDict
#     from sqlalchemy_json import NestedMutableJson, MutableDict, mutable_json_type

#     JsonStringField: t.Type[MutableDict] = mutable_json_type(dbtype = JsonString, nested=True)
#     JsonStringNestedField: t.Type[NestedMutableJson] = mutable_json_type(dbtype = JsonString, nested = True) # type: ignore


#     JSONBField: t.Type[MutableDict] = mutable_json_type(dbtype = JSONB)
#     JSONBNestedField: t.Type[NestedMutableJson] = mutable_json_type(dbtype = JSONB, nested = True) # type: ignore


# for reflection
ischema_names['file'] = FileField
ischema_names['jsonstring'] = JsonString
ischema_names['serializedbinary'] = SerializedBinary

# Try to import pgvector
try:
    from pgvector.sqlalchemy import Vector
    _vector_available = True

except ImportError:
    try:
        import numpy as np

        def from_db(value):
            # could be ndarray if already cast by lower-level driver
            if value is None or isinstance(value, np.ndarray):
                return value
            return np.array(value[1:-1].split(','), dtype=np.float32)
            
        def to_db(value, dim=None):
            if value is None: return value
            if isinstance(value, np.ndarray):
                if value.ndim != 1:
                    raise ValueError('expected ndim to be 1')
                if not np.issubdtype(value.dtype, np.integer) and not np.issubdtype(value.dtype, np.floating):
                    raise ValueError('dtype must be numeric')
                value = value.tolist()
            if dim is not None and len(value) != dim:
                raise ValueError('expected %d dimensions, not %d' % (dim, len(value)))
            return '[' + ','.join([str(float(v)) for v in value]) + ']'


        class Vector(types.UserDefinedType):
            cache_ok = True
            _string = types.String()

            def __init__(self, dim=None):
                super(types.UserDefinedType, self).__init__()
                self.dim = dim

            def get_col_spec(self, **kw):
                return "VECTOR" if self.dim is None else "VECTOR(%d)" % self.dim

            def bind_processor(self, dialect):
                def process(value):
                    return to_db(value, self.dim)
                return process

            def literal_processor(self, dialect):
                string_literal_processor = self._string._cached_literal_processor(dialect)
                def process(value):
                    return string_literal_processor(to_db(value, self.dim))
                return process

            def result_processor(self, dialect, coltype):
                def process(value):
                    return from_db(value)
                return process

            class comparator_factory(types.UserDefinedType.Comparator):
                def l2_distance(self, other):
                    return self.op('<->', return_type=types.Float)(other)

                def max_inner_product(self, other):
                    return self.op('<#>', return_type=types.Float)(other)

                def cosine_distance(self, other):
                    return self.op('<=>', return_type=types.Float)(other)

        ischema_names['vector'] = Vector
        _vector_available = True
    except ImportError:
        _vector_available = False
        class Vector(types.UserDefinedType):
            cache_ok = True
            _string = types.String()

            def __init__(self, dim=None):
                raise NotImplementedError('Vector is not available for without numpy')

# if _vector_available:

class HalfVector(Vector):
    def get_col_spec(self, **kw):
        return "HALFVEC" if self.dim is None else f"HALFVEC({self.dim})"

ischema_names['halfvec'] = HalfVector

# from pgvector.asyncpg import Vector



from __future__ import annotations

"""
Resolver for Pydantic v1/v2 imports with additional helpers
"""
import typing
import functools
from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo
from pydantic.types import ByteSize
from lzl.require import resolve_missing

# Handle v1/v2 of pydantic
try:
    from pydantic import field_validator
    from pydantic import model_validator
    from pydantic import validator as _validator
    from pydantic import ConfigDict
    from pydantic import PrivateAttr

    PYDANTIC_VERSION = 2

    def root_validator(*args, **kwargs):
        """
        v1 Compatible root validator
        """
        def decorator(func):
            _pre_kw = kwargs.pop('pre', None)
            kwargs['mode'] = 'before' if _pre_kw is True else ('after' if _pre_kw is False else kwargs.get('mode', 'wrap'))
            return model_validator(*args, **kwargs)(func)
        return decorator

    def pre_root_validator(*args, **kwargs):
        def decorator(func):
            return model_validator(*args, mode='before', **kwargs)(func)
        return decorator
    
    def validator(*args, **kwargs):
        def decorator(func):
            return _validator(*args, **kwargs)(classmethod(func))
        return decorator

except ImportError:
    from pydantic import root_validator, validator

    ConfigDict = typing.Dict[str, typing.Any]

    PYDANTIC_VERSION = 1

    def pre_root_validator(*args, **kwargs):
        def decorator(func):
            return root_validator(*args, pre=True, **kwargs)(func)
        return decorator

    def field_validator(*args, **kwargs):
        def decorator(func):
            return validator(*args, **kwargs)(func)
        return decorator
    
    def model_validator(*args, **kwargs):
        def decorator(func):
            return root_validator(*args, **kwargs)(func)
        return decorator
    
    PrivateAttr = functools.partial(Field, exclude = True)

try:
    from pydantic_settings import BaseSettings

except ImportError:
    if PYDANTIC_VERSION == 2:
        resolve_missing('pydantic-settings', required = True)
        from pydantic_settings import BaseSettings
    else:
        from pydantic import BaseSettings



def get_pydantic_dict(model: typing.Union[BaseModel, BaseSettings], **kwargs) -> typing.Dict[str, typing.Any]:
    """
    Get a dict from a pydantic model
    """
    if kwargs: kwargs = {k:v for k,v in kwargs.items() if v is not None}
    if hasattr(model, 'model_dump'):
        return model.model_dump(**kwargs)
    return model.dict(**kwargs)

def get_pydantic_fields_dict(model: typing.Type[typing.Union[BaseModel, BaseSettings]]) -> typing.Dict[str, FieldInfo]:
    """
    Get a dict of fields from a pydantic model
    """
    if hasattr(model, 'model_fields'):
        return model.model_fields
    return model.__fields__

def get_pydantic_field_names(model: typing.Type[typing.Union[BaseModel, BaseSettings]]) -> typing.List[str]:
    """
    Get a list of field names from a pydantic model
    """
    return list(get_pydantic_fields_dict(model).keys())

def get_pydantic_fields(model: typing.Type[typing.Union[BaseModel, BaseSettings]]) -> typing.List[FieldInfo]:
    """
    Get a list of fields from a pydantic model
    """
    return list(get_pydantic_fields_dict(model).values())

def pydantic_parse_obj(model: typing.Type[typing.Union[BaseModel, BaseSettings]], obj: typing.Any, **kwargs) -> typing.Union[BaseModel, BaseSettings]:
    """
    Parse an object into a pydantic model
    """
    if hasattr(model, 'model_validate'):
        return model.model_validate(obj, **kwargs)
    return model.parse_obj(obj)

def get_pydantic_schema(model: typing.Type[typing.Union[BaseModel, BaseSettings]], **kwargs) -> typing.Dict[str, typing.Any]:
    """
    Get a pydantic schema
    """
    if hasattr(model, 'model_json_schema'):
        return model.model_json_schema(**kwargs)
    return model.schema(**kwargs)


def get_schema_extra(schema: typing.Dict[str, typing.Any], _):
    """
    Helper to get the extra schema
    """
    props = {
        k: v
        for k, v in schema.get('properties', {}).items()
        if not v.get("hidden", False) and not k.startswith('_')
    }
    schema["properties"] = props



from __future__ import annotations


from lzl.api import aiohttpx
from ...types.client import BaseOAuth2Client, eproperty, jwt
from ...types import errors
from .config import GoogleOAuth2Config
from typing import Any, Dict, Optional, Union, Tuple, TYPE_CHECKING, List, Type, Annotated

if TYPE_CHECKING:
    from ...types.user import OAuth2User

class GoogleOAuth2Client(BaseOAuth2Client):
    """
    Google Auth Client
    """
    name: Optional[str] = 'google'

    if TYPE_CHECKING:
        config: GoogleOAuth2Config
    

    @eproperty
    def config_class(self) -> Type['GoogleOAuth2Config']:
        """
        Returns the Config Class
        """
        return GoogleOAuth2Config
    
    def fetch_token_info(
        self,
        token: str,
        kind: str = 'access_token',
        **kwargs
    ) -> Optional[Dict[str, Any]]:
        """
        Fetches the token info
        """
        response = self.client.get(self.config.token_info_endpoint, params = {kind: token})
        response.raise_for_status()
        return response.json()
    
    async def afetch_token_info(
        self,
        token: str,
        kind: str = 'access_token',
        **kwargs
    ) -> Optional[Dict[str, Any]]:
        """
        Fetches the token info
        """
        response = await self.client.async_get(self.config.token_info_endpoint, params = {kind: token})
        try:
            response.raise_for_status()
        except aiohttpx.HTTPStatusError as e:
            raise errors.InvalidAuthorizationError(f'Error Fetching Token Info: {e.response.text}') from e
        except Exception as e:
            raise errors.InvalidAuthorizationError('Error Fetching Token Info', e) from e
        return response.json()

    def parse_jwt_token(
        self,
        jwt_token: str,
        kind: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Parses the jwt token
        """
        # Posssible bug with the jwt?
        if kind == 'access_token': 
            return self.fetch_token_info(jwt_token, kind = kind)
        return jwt.decode(
            jwt_token,
            "",
            # signing_key.key, 
            algorithms=['RS256'],
            issuer = self.config.issuer_url,
            options = {
                'verify_signature': False,
                'verify_exp': True,
                'verify_aud': False,
            }
        )
        # return super().parse_jwt_token(jwt_token, kind = kind)

    
    def parse_authorization_token(
        self,
        authorization: str,
        audiences: Optional[List[str]] = None,
        verify_exp: bool = True,
    ) -> Optional[Dict[str, Any]]:
        """
        Parses the authorization token
        """
        # Posssible bug with the jwt?
        return jwt.decode(
            authorization, 
            "", 
            algorithms = ['RS256'],
            audience = audiences,
            issuer = self.config.jwt_issuer_url,
            options={
                'verify_signature': False,
                'verify_exp': verify_exp,
                'verify_aud': bool(audiences),
            }
        )
    

    async def adecode_jwt_token(
        self,
        token: str,
        options: Optional[Dict[str, bool]] = None,
        disable_signing_key: Optional[bool] = None,
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Decodes the jwt token
        """
        if token.startswith('ya29.'):
            return await self.afetch_token_info(
                token, 
                kind = 'access_token',
            )
        return await super().adecode_jwt_token(token, options = options, disable_signing_key = disable_signing_key, **kwargs)
    
    async def set_user_data(self, user: 'OAuth2User', **kwargs) -> None:
        """
        Sets the user data
        """
        if not user.token: return
        if user.token.access_token:
            if user.token.access_token.scope:
                user.scopes = user.token.access_token.scope.split(' ')

        if user.token.id_token:
            extra_data = {
                'picture': user.token.id_token.picture,
                'at_hash': user.token.id_token.at_hash,
                'name': user.token.id_token.name,
                'given_name': user.token.id_token.given_name,
                'family_name': user.token.id_token.family_name,
                'email': user.token.id_token.email,
                'email_verified': user.token.id_token.email_verified,
            }
            extra_data = {k: v for k, v in extra_data.items() if v}
            if not user.data: user.data = {}
            user.data.update(extra_data)
        return user


    # async def avalidate_user(
    #     self,
    #     user: 'OAuth2User',
    #     **kwargs,
    # ) -> 'OAuth2User':
    #     """
    #     Validates the user
    #     """
    #     pass

    # async def aretrieve_user_info(
    #     self,
    #     user: 'OAuth2User',
    #     **kwargs,
    # ) -> Optional[Dict[str, Any]]:
    #     """
    #     Retrieves the user info
    #     """
    #     pass





        


from __future__ import annotations

"""
aiohttpx module. Formally from `growthengineai/aiohttpx`. No longer maintained there.

A Submodule for httpx that combines
the httpx.AsyncClient with the httpx.Client
to allow both async and sync requests

Usage:

```python

from lzl.api import aiohttpx

with aiohttpx.Client() as client:
    
    # Make an Async GET request
    response = await client.async_get("https://httpbin.org/get")
    print(response.json())

    # Make a Sync GET request
    response = client.get("https://httpbin.org/get")
    print(response.json())


base_url = "https://www.google.com"

async with aiohttpx.ProxyClient(base_url = base_url) as client:
    # Make an Async GET request
    response = await client.async_get(
        "/search", 
        params = {"q": "httpx"},
        soup_enabled = True
    )
    print(response.soup)
    print(response.soup.title.text)

    # Make a Sync GET request
    response = client.get(
        "/search", 
        params = {"q": "httpx"},
        soup_enabled = True
    )
    print(response.soup)
    print(response.soup.title.text)

```
"""


# import top level classes from httpx
# to allow for easy access
from lzl.load import lazy_load
h = lazy_load('httpx', install_missing = True)
h.__load__()

from httpx._api import delete, get, head, options, patch, post, put, request, stream
from httpx._auth import Auth, BasicAuth, DigestAuth
from httpx._client import USE_CLIENT_DEFAULT
from httpx._client import AsyncClient as hAsyncClient
from httpx._client import Client as hClient

from httpx._config import Limits, Proxy, Timeout, create_ssl_context
from httpx._content import ByteStream
from httpx._exceptions import (
    CloseError,
    ConnectError,
    ConnectTimeout,
    CookieConflict,
    DecodingError,
    HTTPError,
    HTTPStatusError,
    InvalidURL,
    LocalProtocolError,
    NetworkError,
    PoolTimeout,
    ProtocolError,
    ProxyError,
    ReadError,
    ReadTimeout,
    RemoteProtocolError,
    RequestError,
    RequestNotRead,
    ResponseNotRead,
    StreamClosed,
    StreamConsumed,
    StreamError,
    TimeoutException,
    TooManyRedirects,
    TransportError,
    UnsupportedProtocol,
    WriteError,
    WriteTimeout,
)

from httpx._models import Cookies, Headers, Request
from .client import Response
from httpx._status_codes import codes
from httpx._types import AsyncByteStream, SyncByteStream
try:
    from httpx._urls import URL, QueryParams
except ImportError:
    from httpx._models import URL, QueryParams
from urllib.parse import (
    urlparse, 
    urlunparse, 
    urljoin, 
    urldefrag,
    urlsplit, 
    urlunsplit, 
    urlencode, 
    parse_qs,
    parse_qsl, 
    quote, 
    quote_plus, 
    quote_from_bytes,
    unquote, 
    unquote_plus,
    unquote_to_bytes,
)

from .api import (
    async_delete, 
    async_get, 
    async_head, 
    async_options, 
    async_patch, 
    async_post, 
    async_put, 
    async_request, 
    async_stream,
    acreate_stream, 
    astream,
    aget, 
    ahead, 
    aoptions, 
    apatch, 
    apost, 
    aput, 
    arequest, 
    adelete,
    stream_iterator,
    astream_iterator,
)
from .client import Client
from .presets import PresetConfig, get_preset
from .types.params import ClientParams
from .utils.logs import mute_httpx_logger
from .utils.helpers import raise_for_status, http_retry_wrapper


from __future__ import annotations

"""
Base Hatchet Client with Modifications

Patched: 2024-08-08 19:02:25
Version: 0.32.0
Last Version: 0.31.0
"""

import os
import signal
import asyncio
import functools
import contextvars
import grpc
from threading import Thread
from concurrent.futures import Future, ThreadPoolExecutor
from hatchet_sdk.loader import ClientConfig
from hatchet_sdk.worker import (
    Worker as BaseWorker, 
    WorkerStatus, 
    errorWithTraceback, 
    copy_context_vars,
    wr, sr,
    new_dispatcher, 
    new_admin,
    new_listener,
    PooledWorkflowRunListener,
    GetActionListenerRequest,
    WorkerContext
)
from hatchet_sdk.client import new_client_raw
from threading import Thread, current_thread
from hatchet_sdk.clients.dispatcher import Action, ActionListenerImpl
from hatchet_sdk.dispatcher_pb2 import (
    GROUP_KEY_EVENT_TYPE_COMPLETED,
    GROUP_KEY_EVENT_TYPE_FAILED,
    GROUP_KEY_EVENT_TYPE_STARTED,
    STEP_EVENT_TYPE_COMPLETED,
    STEP_EVENT_TYPE_FAILED,
    STEP_EVENT_TYPE_STARTED,
    ActionType,
    GroupKeyActionEvent,
    GroupKeyActionEventType,
    StepActionEvent,
    StepActionEventType,
    WorkerUnsubscribeRequest,
)
from hatchet_sdk.logger import logger
from hatchet_sdk.metadata import get_metadata
from typing import Any, Dict, Callable, Type, Optional, TYPE_CHECKING
from .context import Context, ContextT
from .utils import new_session, json_serializer
from lzo.utils.logs import logger as _logger

if TYPE_CHECKING:
    from .session import HatchetSession

class Worker(BaseWorker):
    """
    The Worker Object
    """
    SIGNALS = [signal.SIGINT, signal.SIGTERM] if os.name != "nt" else [signal.SIGTERM]

    

    def __init__(
        self,
        name: str,
        max_runs: int | None = None,
        debug=False,
        handle_kill=True,
        config: ClientConfig = {},
        labels: dict[str, str | int] = {},
        session: Optional['HatchetSession'] = None,
        context_class: Type[Context] = Context,

    ):
        if debug:
            logger.warn(
                "debug on worker is deprecated and will be removed in a future release, please set debug on the Hatchet client instead"
            )

        # We store the config so we can dynamically create clients for the dispatcher client.
        self.config = config
        self.session = session
        self.client = self.session.client if self.session else new_session(config)
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._context_cls = context_class
        self.main_task = None

        self.name = self.client.config.namespace + name
        self.max_runs = max_runs
        self.tasks: Dict[str, asyncio.Task] = {}  # Store run ids and futures
        self.contexts: Dict[str, Context] = {}  # Store run ids and contexts
        self.action_registry: dict[str, Callable[..., Any]] = {}
        self.listener: ActionListenerImpl = None

        # The thread pool is used for synchronous functions which need to run concurrently
        self.thread_pool = ThreadPoolExecutor(max_workers=max_runs)
        self.threads: Dict[str, Thread] = {}  # Store run ids and threads

        self.killing = False
        self.handle_kill = handle_kill

        self._status = WorkerStatus.INITIALIZED

        self.worker_context = WorkerContext(labels=labels, config=config)


    async def handle_start_step_run(self, action: Action):
        logger.debug(f"Starting step run {action.step_run_id}")

        action_name = action.action_id
        context = self._context_cls(
            action,
            self.dispatcher_client,
            self.admin_client,
            self.client.event,
            self.client.workflow_listener,
            self.workflow_run_event_listener,
            self.worker_context,
            self.client.config.namespace,
        )
        self.contexts[action.step_run_id] = context

        # Find the corresponding action function from the registry
        action_func = self.action_registry.get(action_name)

        if action_func:
            # send an event that the step run has started
            try:
                event = self.get_step_action_event(action, STEP_EVENT_TYPE_STARTED)

                # Send the action event to the dispatcher
                asyncio.create_task(
                    self.dispatcher_client.send_step_action_event(event)
                )
            except Exception as e:
                logger.error(f"Could not send action event: {e}")

            task = self.loop.create_task(
                self.async_wrapped_action_func(
                    context, action_func, action, action.step_run_id
                )
            )

            task.add_done_callback(self.step_run_callback(action))
            self.tasks[action.step_run_id] = task

            try:
                await task
            except Exception as e:
                # do nothing, this should be caught in the callback
                pass

        logger.debug(f"Finished step run {action.step_run_id}")


    async def handle_start_group_key_run(self, action: Action):
        action_name = action.action_id
        context = self._context_cls(
            action,
            self.dispatcher_client,
            self.admin_client,
            self.client.event,
            self.client.workflow_listener,
            self.workflow_run_event_listener,
            self.worker_context,
            self.client.config.namespace,
        )
        self.contexts[action.get_group_key_run_id] = context

        # Find the corresponding action function from the registry
        action_func = self.action_registry.get(action_name)

        if action_func:
            # send an event that the group key run has started
            try:
                event = self.get_group_key_action_event(
                    action, GROUP_KEY_EVENT_TYPE_STARTED
                )

                # Send the action event to the dispatcher
                asyncio.create_task(
                    self.dispatcher_client.send_group_key_action_event(event)
                )
            except Exception as e:
                if "]:" in str(e) and "- 40" in str(e):
                    logger.error(f"[{action.step_run_id} - {action.job_name}] Error in action: {e}")
                    raise e
                logger.error(f"Could not send action event: {e}")

            task = self.loop.create_task(
                self.async_wrapped_action_func(
                    context, action_func, action, action.get_group_key_run_id
                )
            )

            task.add_done_callback(self.group_key_run_callback(action))
            self.tasks[action.get_group_key_run_id] = task

            try:
                await task
            except Exception as e:
                # do nothing, this should be caught in the callback
                pass


    async def async_wrapped_action_func(
        self, context: Context, action_func, action: Action, run_id: str
    ):
        wr.set(context.workflow_run_id())
        sr.set(context.step_run_id)

        try:
            if action_func._is_coroutine:
                return await action_func(context)
            else:
                pfunc = functools.partial(
                    # we must copy the context vars to the new thread, as only asyncio natively supports
                    # contextvars
                    copy_context_vars,
                    contextvars.copy_context().items(),
                    self.thread_action_func,
                    context,
                    action_func,
                    action,
                )
                res = await self.loop.run_in_executor(self.thread_pool, pfunc)

                return res
        except Exception as e:
            if "]:" in str(e) and "- 40" in str(e):
                logger.error(f"[{action.step_run_id} - {action.job_name}] Error in action: {e}")
                raise e
            logger.error(errorWithTraceback(f"Could not execute action: {e}", e))
            raise e
        finally:
            self.cleanup_run_id(run_id)


    def get_step_action_finished_event(
        self, action: Action, output: Any
    ) -> StepActionEvent:
        try:
            event = self.get_step_action_event(action, STEP_EVENT_TYPE_COMPLETED)
        except Exception as e:
            logger.error(f"Could not create action finished event: {e}")
            raise e

        output_bytes = ""

        if output is not None:
            output_bytes = json_serializer.dumps(output)

        event.eventPayload = output_bytes

        return event



    # @property
    # def loop(self) -> asyncio.AbstractEventLoop:
    #     """
    #     Returns the event loop
    #     """
    #     if not self._loop: 
    #         try:
    #             self._loop = asyncio.get_running_loop()
    #         except RuntimeError as e:
    #             logger.error(f"Error getting running loop: {e}")
    #             try:
    #                 self._loop = asyncio.get_event_loop()
    #             except RuntimeError as e:
    #                 logger.error(f"Error getting event loop: {e}")
    #                 self._loop = asyncio.new_event_loop()
    #                 asyncio.set_event_loop(self._loop)
    #     return self._loop
    
    # @loop.setter
    # def loop(self, value: asyncio.AbstractEventLoop):
    #     """
    #     Sets the loop
    #     """
    #     self._loop = value

    # def run(self):
    #     """
    #     Run function
    #     """
    #     self.main_task = self.loop.create_task(self.async_run)
    #     try:
    #         self.loop.run_until_complete(self.main_task)
    #     except asyncio.CancelledError:  # pragma: no cover
    #         # happens on shutdown, fine
    #         pass
    #     finally:
    #         self.loop.run_until_complete(self.exit_gracefully)

    async def watch_for_termination(self):
        """
        Watches for termination
        """
        self.event = asyncio.Event()
        
        try:
            for signum in self.SIGNALS: self.loop.add_signal_handler(signum, self.event.set)
            await self.event.wait()
        # except asyncio.CancelledError:  # pragma: no cover
            # happens on shutdown, fine
            # self.event.set()
        except Exception as e:
            _logger.trace('Error in Watching', e)
            # self.event.set()
        finally:
            self.event.set()
            logger.warning(f'Exiting!')
            self.loop.run_until_complete(self.exit_gracefully)

    def run_watcher_event(self):
        """
        Starts the watcher event
        """
        asyncio.create_task(self.watch_for_termination())
    
    async def async_run(self, **kwargs):
        """
        Async Function to run the worker
        """
        self.event = asyncio.Event()
        for signum in self.SIGNALS: self.loop.add_signal_handler(signum, self.event.set)
        task = None
        try:
            task = self.loop.create_task(self.async_start)
            await self.event.wait()
        except Exception as e:
            logger.error(f'Shutdown: {e}')
            raise e
        finally:
            self.event.set()
            if task is not None:
                if not task.done(): task.cancel()
            if self.main_task: self.main_task.cancel()
            await asyncio.wait_for(asyncio.gather(task), timeout = 5.0)







from __future__ import annotations

"""High-level IO façade.

This module publicly re-exports the primary IO helpers used across LazyOps.
Keeping the concise surface area documented in a single place makes it easier
for downstream projects – and Mintlify generated documentation – to link to the
canonical entry points without needing to traverse the underlying package
structure.
"""

import typing as t

from .file import File, FileLike, PathLike

if t.TYPE_CHECKING:
    from .ser import (
        SerT, 
        JsonSerializer,
        PickleSerializer,
    )
    from .compression import CompressionT
    from .persistence import PersistentDict, TemporaryData

__all__ = ["File", "FileLike", "PathLike"]



from __future__ import annotations


"""
Middleware Utilities
"""

import os
import sys
import typing
from uuid import uuid4
from pathlib import Path
from starlette.datastructures import MutableHeaders, Secret
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Message, Receive, Scope, Send
from typing import Any, Dict, List, Optional, Union, Type, TYPE_CHECKING, Callable, Awaitable
from lzl.logging import logger
from lzl.types import Literal

if TYPE_CHECKING:
    from lzl.io import PersistentDict
    from lzl.pool import ThreadPool

try:
    import xxhash

    def hash_func(data: str) -> str:
        """
        Returns the hash of the data
        """
        return xxhash.xxh128_hexdigest(data)

except ImportError:
    import hashlib

    def hash_func(data: str) -> str:
        """
        Returns the hash of the data
        """
        return hashlib.sha256(data.encode()).hexdigest()

def create_session_key(*keys: str) -> str:
    """
    Creates a session key
    """
    joined_keys = ':'.join(keys)
    return hash_func(joined_keys)

def get_pdict(
    base_key: str,
    redis_url: Optional[str] = None,
    serializer: Optional[str] = 'json',
    serializer_kwargs: Optional[Dict[str, Any]] = None,
    expiration: Optional[int] = None,
    allow_local_fallback: bool = True,
    local_cache_file: Optional[Path] = None,
    session_name: Optional[str] = 'fastapi.session',
    **kwargs,
) -> 'PersistentDict':
    """
    Returns the Persistent Dict for data
    """
    try:
        from kvdb import KVDBClient
        sess = KVDBClient.get_session(
            name = session_name,
            url = redis_url,
        )
        from lzo.utils.helpers.base import fail_after
        fail_func = fail_after(5)(sess.ping)
        try:
            # with fail_after(5):
            #     sess.ping()
            fail_func()
        except Exception as e:
            if not allow_local_fallback: raise ValueError(f'KVDB Sesssion is not available and local fallback is not allowed: {e}') from e
            logger.warning(f'KVDB Sesssion is not available and local fallback is not allowed: {e}')
            sess = None
    except ImportError as e:
        if not allow_local_fallback: raise ImportError('Persistent Dict is not available and local fallback is not allowed') from e
    
    if sess is not None:
        return sess.create_persistence(
            base_key = base_key,
            serializer = serializer,
            expiration = expiration,
            serializer_kwargs = serializer_kwargs,
            **kwargs,
        )
    from lzl.io.persistence import PersistentDict
    return PersistentDict(
        base_key = base_key,
        serializer = serializer,
        expiration = expiration,
        serializer_kwargs = serializer_kwargs,
        file_path = local_cache_file,
        **kwargs,
    )





from __future__ import annotations

"""
Env Var Helpers
"""

import os
import re
import json
import subprocess
import warnings
warnings.filterwarnings("ignore", message = "invalid escape sequence")
warnings.filterwarnings("ignore", category=SyntaxWarning)

from pathlib import Path
from .formatting import build_dict_from_str
from typing import Any, Dict, List, Optional, Tuple, Union, Type, TypeVar, TYPE_CHECKING

EnvT = TypeVar('EnvT')

def parse_from_envvar(
    var: str,
    default: Optional[Any] = None,
    _type: Optional[Type[EnvT]] = str,
) -> EnvT:
    """
    Parse a value from an environment variable
    """
    val = os.getenv(var)
    if val is None: return default

    # Try to evaluate the value
    if _type is None: 
        # print('No type: ', val)
        return val
        # if val.startswith(('{', '[')):
        #     return json.loads(val)
        # return build_dict_from_str(val) if val.startswith('{') else val.split(',')
    if _type in {list, dict}:
        if val.startswith(('{', '[')):
            return json.loads(val)
        return build_dict_from_str(val) if _type is dict else val.split(',')

    if _type is Path:
        return Path(val)
    
    if _type is bool:
        return val.lower() in {'true', '1', 't', 'y', 'yes'}
    
    return _type(val)


def parse_envvars_from_text(
    text: str,
    values: Optional[Dict[str, Any]] = None,
    envvar_prefix: Optional[str] = 'env/',
) -> Tuple[str, Dict[str, Any]]:
    """
    Parse values from a text block

    Returns:
        Tuple[str, Dict[str, Any]]: The text with the envvars replaced and the parsed values
    """

    # Create a pattern that would match env/ENVVAR_NAME and capture the ENVVAR_NAME
    _prefix = envvar_prefix.replace('/', '\/')
    pattern = re.compile(rf'({_prefix}\w+)')

    # Find all
    values = values or {}
    matches = pattern.findall(text)
    for match in matches:
        # var = match[0]
        envvar = match.replace(envvar_prefix, '')
        # print(match, envvar)
        _default = values.get(envvar)
        if isinstance(_default, type):
            _type = _default
            _default = None
        else:
            _type = type(_default) if _default is not None else None
        val = parse_from_envvar(envvar, default=_default, _type=_type)
        # if val is not None:
        values[envvar] = val
        # print(match, val, values)
        text = text.replace(match, val if val is not None else '')

    return text, values

# _envvar_pattern = re.compile(r"\$\{([A-Za-z_][A-Za-z0-9_]*)\}")
# _envvar_pattern = re.compile(r"\$\{([A-Za-z_][A-Za-z0-9_]*(?:\.[A-Za-z_][A-Za-z0-9_]*)*(?:\:[A-Za-z_][A-Za-z0-9_]*)*)\}")
# _envvar_pattern = re.compile(r"\$\{([A-Za-z_][\w\.]*(?:\:[A-Za-z_][\w]*)*)\}")
_envvar_pattern = re.compile(r"\$\{([^}]*)\}")
_envkey_getters = {
    'HOST': (lambda: os.uname().nodename), 
    'HOSTNAME': (lambda: os.uname().nodename),
}

def dummy_sample(t: str, *args) -> str:
    return f'{t}-1'

def get_cmd_output(cmd: str, *args) -> str:
    """
    Runs a command and returns the output
    """
    cmd = cmd.lstrip('$(').rstrip(')')
    with os.popen(cmd) as pipe:
        return pipe.read().strip()

        
def replace_envvars_in_text(
    text: str,
    enable_function_calls: Optional[bool | str] = 'auto',
    enable_shell_commands: Optional[bool | str] = 'auto',
) -> str:
    """
    Replaces all envvars in a text with their values

    All env vars need to be wrapped in ${} to be replaced.

    If `enable_function_calls` is True, then any envvars that are prefixed with
    `func:` will be treated as a function call and the result will be used as the value.

    If `enable_shell_commands` is True, then any envvars that are prefixed with
    `cmd:` will be treated as a shell command and the result will be used as the value.

    Example:
        >>> replace_envvars_in_text("Hello, ${NAME}!")
        "Hello, John!"

        >>> replace_envvars_in_text("Hello, ${func:get_name:John}!")
        "Hello, John!"

        >>> replace_envvars_in_text("Hello, ${cmd:$(echo 'John')}!")
        "Hello, John!"

    """
    from lzl.load import lazy_import
    if enable_function_calls == 'auto': enable_function_calls = '${func:' in text
    if enable_shell_commands == 'auto': enable_shell_commands = '${cmd:' in text
    for match in _envvar_pattern.finditer(text):
        envvar = match.group(1)
        _default = ''
        # print(envvar)
        if enable_function_calls and envvar.startswith('func:'):
            parts = envvar.split(':')
            args = parts[2:] or []
            func = lazy_import(parts[1])
            value = func(*args)
        elif enable_shell_commands and envvar.startswith('cmd:'):
            parts = envvar.split(':')
            args = parts[2:] or []
            value = get_cmd_output(parts[1], *args)
        elif envvar in _envkey_getters: value = _envkey_getters[envvar]()
        else: 
            if ':-' in envvar:
                envvar, _default = envvar.split(':-')
            value = os.getenv(envvar)
        value = value or _default
        text = text.replace(match.group(0), value)
    return text


def load_env_vars(
    env_file: str,
    categories: Optional[Union[str, List[str]]] = None,
    target_app: Optional[str] = None,
    included_apps: Optional[Union[str, List[str]]] = None,
    excluded_apps: Optional[Union[str, List[str]]] = None,
    included_vars: Optional[Union[str, List[str]]] = None,
    excluded_vars: Optional[Union[str, List[str]]] = None,
    set_as_env: Optional[bool] = False,
    overwrite: Optional[bool] = True,
) -> Dict[str, str]:
    """
    Load Environment Variables from a YAML file

    """
    import yaml
    configs: List[Dict[str, Union[str, Dict[str, Any], List[str]]]] = yaml.safe_load(
        Path(env_file).read_text()
    )
    if included_vars and not isinstance(included_vars, list):
        included_vars = [included_vars]
    if excluded_vars and not isinstance(excluded_vars, list):
        excluded_vars = [excluded_vars]
    if included_apps and not isinstance(included_apps, list):
        included_apps = [included_apps]
    if excluded_apps and not isinstance(excluded_apps, list):
        excluded_apps = [excluded_apps]
    if categories and not isinstance(categories, list):
        categories = [categories]
    
    if target_app:
        if not included_apps: included_apps = []
        included_apps.append(target_app)
    
    envvars: Dict[str, str] = {}
    for config in configs:
        if config.get('disabled', False): continue
        if config.get('category') and categories and \
            all(cat not in config['category'] for cat in categories):
            continue
        if included_vars and config['name'] not in included_vars:
            continue
        if excluded_vars and config['name'] in excluded_vars:
            continue
        if config.get('apps'):
            if included_apps and all(app not in included_apps for app in config['apps']): 
                continue
            if excluded_apps and any(app in excluded_apps for app in config['apps']):
                continue
        for key, value in config['envs'].items():
            if not value: continue
            if isinstance(value, dict) and target_app:
                value = value.get(target_app)
            if isinstance(value, list):
                value = f'[{", ".join(value)}]'
            elif isinstance(value, dict):
                value = json.dumps(value)
            elif not isinstance(value, str):
                value = str(value)
            envvars[key] = value
            if set_as_env and (overwrite or key not in os.environ):
                os.environ[key] = value
    return envvars


_default_envkeys = [
    'IN_CI_ENV',
    'DOCKER_CI',
    'BUILDER_CI',
    'APP_ENV=CI',
    'BUILD_ENV=CI',
]

_is_in_ci: Optional[bool] = None

def is_in_ci_env(
    envkeys: Optional[List[str]] = _default_envkeys,
    enable_global_check: bool = True,
) -> bool:
    """
    Checks if the code is running in a CI environment
    """
    global _is_in_ci
    if _is_in_ci is not None and enable_global_check: 
        return _is_in_ci
    
    def _run_check():
        for key in envkeys:
            if '=' in key:
                key, value = key.split('=', 1)
                if os.getenv(key, '') == value:
                    return True
            elif os.getenv(key, '').lower() in {'true', '1', 't', 'y', 'yes'}:
                return True
        return False
    
    _is_in_ci = _run_check()
    return _is_in_ci


from __future__ import annotations

"""
Kinde OAuth2 Client
"""

from ...types.client import BaseOAuth2Client, eproperty
from typing import Any, Dict, Optional, Union, Tuple, TYPE_CHECKING, List, Type, Annotated
from .config import KindeOAuth2Config
from .types import UserInfo

if TYPE_CHECKING:
    from lzl.io import PersistentDict
    from lzl.io.ser import JsonSerializer
    from ...types.user import OAuth2User

class KindeOAuth2Client(BaseOAuth2Client):
    """
    Kinde OAuth2 Client
    """
    name: Optional[str] = 'kinde'

    if TYPE_CHECKING:
        config: KindeOAuth2Config

    
    @property
    def user_data(self) -> 'PersistentDict[str, UserInfo]':
        """
        Returns the user info data
        """
        if 'user_data' not in self._extra:
            self._extra['user_data'] = self.data.get_child('users', expiration = 300)
        return self._extra['user_data']


    @eproperty
    def config_class(self) -> Type['KindeOAuth2Config']:
        """
        Returns the Config Class
        """
        return KindeOAuth2Config

    @eproperty
    def serializer(self) -> 'JsonSerializer':
        """
        Returns the serializer
        """
        from lzl.io.ser import get_serializer
        return get_serializer('json', disable_nested_values = True)


    async def app_retrieve_existing_callbacks(
        self,
        **kwargs,
    ) -> Optional[List[str]]:
        """
        Retrieves the existing app callbacks
        """
        response = await self.m2m_client.async_get(
            f'/api/v1/applications/{self.config.client_id}/auth_redirect_urls',
            headers = {
                'Accept': 'application/json',
            }
        )
        response.raise_for_status()
        data: Dict[str, List[str]] = response.json()
        # logger.info(data)
        return data.get('redirect_urls', [])


    async def app_add_authorized_callbacks(
        self,
        callback_urls: List[str],
        **kwargs,
    ) -> None:
        """
        Adds the authorized callbacks
        """
        response = await self.m2m_client.async_post(
            f'/api/v1/applications/{self.config.client_id}/auth_redirect_urls',
            headers = {
                'Accept': 'application/json',
                'Content-Type': 'application/json; charset=utf-8',
            },
            json = {
                'urls': callback_urls,
            }
        )
        response.raise_for_status()
        data = response.json()
        data['urls'] = callback_urls
        self.logger.info(data, colored = True, prefix = f'{self.name} - Added {len(callback_urls)} Authorized Callbacks')
        
    
    async def app_retrieve_existing_logouts(
        self,
        **kwargs,
    ) -> Optional[List[str]]:
        """
        Retrieves the existing app logout urls
        """
        response = await self.m2m_client.async_get(
            f'/api/v1/applications/{self.config.client_id}/auth_logout_urls',
            headers = {
                'Accept': 'application/json',
            }
        )
        response.raise_for_status()
        data: Dict[str, List[str]] = response.json()
        return data.get('logout_urls', [])

    async def app_add_authorized_logouts(
        self,
        logout_urls: List[str],
        **kwargs,
    ) -> None:
        """
        Adds the authorized logouts
        """
        response = await self.m2m_client.async_post(
            f'/api/v1/applications/{self.config.client_id}/auth_logout_urls',
            headers = {
                'Accept': 'application/json',
                'Content-Type': 'application/json; charset=utf-8',
            },
            json = {
                'urls': logout_urls,
            }
        )
        response.raise_for_status()
        data = response.json()
        data['urls'] = logout_urls
        self.logger.info(data, colored = True, prefix = f'{self.name} - Added {len(logout_urls)} Authorized Logouts')
    
    async def aretrieve_user_info(
        self,
        # user_id: str,
        user: 'OAuth2User',
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Retrieves the user info
        """
        response = await self.m2m_client.async_get(
            f'/api/v1/user?id={user.provider_id}',
            headers = {
                'Accept': 'application/json',
            }
        )
        response.raise_for_status()
        return response.json()
        # data = response.json()
        # self.logger.info(data)
        # return data

    def map_feature_flags(self, feature_flags: Dict[str, Dict[str, Union[str, bool, dict]]]) -> Dict[str, Any]:
        """
        Maps the feature flags
        """
        result = {}
        for key, flag in feature_flags.items():
            if flag['t'] == 's':
                result[key] = flag['v']
                continue
            if flag['t'] == 'b':
                result[key] = flag['v'] == 'true'
                continue
            if flag['t'] == 'i':
                result[key] = int(flag['v'])
                continue
            if flag['t'] == 'j':
                result[key] = self.serializer.loads(flag['v']) if \
                    isinstance(flag['v'], str) else flag['v']
                continue
        return result

    def map_properties(self, props: Dict[str, Dict[str, Union[str, bool]]]) -> Dict[str, Any]:
        """
        Maps the properties
        """
        result = {}
        for key, value in props.items():
            if value.get('v'):
                if '{' in value['v']:
                    value['v'] = self.serializer.loads(value['v'])
                elif value['v'] in {'true', 'false'}:
                    value['v'] = value['v'] == 'true'
            else:
                value['v'] = None
            result[key] = value
        return result
    
    def serialize_properties(self, props: Dict[str, Any]) -> Dict[str, Any]:
        """
        Serializes the properties
        """
        result = {}
        for key, value in props.items():
            if value is None:
                continue
            if isinstance(value, str):
                result[key] = {'v': value}
                continue
            result[key] = {'v': self.serializer.dumps(value)}
        return result
        
    async def set_user_data(self, user: 'OAuth2User', **kwargs) -> None:
        """
        Sets the user data
        """
        # if user.method not in {'session', 'authorization'}: return
        if not user.token: return
        # self.logger.info(f'Fetching User Data: {user.token}')
        # self.logger.info(f'Setting User Data: {user.token}')
        if user.token.access_token.scp:
            user.scopes = user.token.access_token.scp
        roles = [role['name'] for role in user.token.access_token.get('roles', [])]
        for role in roles:
            user.add_role(role)
        extra_data = {}
        if user.token.access_token:
            extra_data.update({
                'external_org_id': user.token.access_token.get('external_org_id'),
                'org_code': user.token.access_token.get('org_code'),
                'org_name': user.token.access_token.get('org_name'),
                'permissions': user.token.access_token.get('permissions'),
            })
            if feature_flags := user.token.access_token.get('feature_flags'):
                extra_data['feature_flags'] = self.map_feature_flags(feature_flags)
            if organization_properties := user.token.access_token.get('organization_properties'):
                extra_data['organization_properties'] = self.map_properties(organization_properties)
            if user_properties := user.token.access_token.get('user_properties'):
                extra_data['user_properties'] = self.map_properties(user_properties)
        if user.token.id_token:
            extra_data.update({
                'org_codes': user.token.id_token.get('org_codes'),
                'organizations': user.token.id_token.get('organizations'),
                'name': user.token.id_token.get('name'),
                'given_name': user.token.id_token.get('given_name'),
                'family_name': user.token.id_token.get('family_name'),
            })
            if not extra_data.get('feature_flags') and (feature_flags := user.token.id_token.get('feature_flags')):
                extra_data['feature_flags'] = self.map_feature_flags(feature_flags)
            if not extra_data.get('organization_properties') and (organization_properties := user.token.id_token.get('organization_properties')):
                extra_data['organization_properties'] = self.map_properties(organization_properties)
            if not extra_data.get('user_properties') and (user_properties := user.token.id_token.get('user_properties')):
                extra_data['user_properties'] = self.map_properties(user_properties)
        extra_data = {k: v for k, v in extra_data.items() if v}
        if not extra_data: return
        if not user.data: user.data = {}
        user.data.update(extra_data)
        return user

    def update_user_properties(
        self,
        user: 'OAuth2User',
    ):
        """
        Updates the user property
        """
        props = {}
        if user.data.get('organization_properties'):
            props.update(user.data['organization_properties'])
        if user.data.get('user_properties'):
            props.update(user.data['user_properties'])
        if not props: return
        response = self.m2m_client.patch(
            f'/api/v1/users/{user.provider_id}/properties',
            headers = {
                'Accept': 'application/json',
                'Content-Type': 'application/json',
            },
            json = {
                'properties': self.serialize_properties(props),
            }
        )
        response.raise_for_status()
        data = response.json()
        self.logger.info(data, colored = True, prefix = f'{self.name} - Updated User Properties')
        return user

    async def aupdate_user_properties(
        self,
        user: 'OAuth2User',
    ):
        """
        Updates the user property
        """
        props = {}
        if user.data.get('organization_properties'):
            props.update(user.data['organization_properties'])
        if user.data.get('user_properties'):
            props.update(user.data['user_properties'])
        if not props: return
        response = await self.m2m_client.async_patch(
            f'/api/v1/users/{user.provider_id}/properties',
            headers = {
                'Accept': 'application/json',
                'Content-Type': 'application/json',
            },
            json = {
                'properties': self.serialize_properties(props),
            }
        )
        response.raise_for_status()
        data = response.json()
        self.logger.info(data, colored = True, prefix = f'{self.name} - Updated User Properties')
        return user


from __future__ import annotations

"""
TemporalIO Telemetry Config
"""

import os
import json
import typing as t
from lzl import load
from datetime import timedelta
from pydantic.types import Json
from lzo.types import BaseSettings, model_validator, Field
from lzl.types import eproperty, create_alias_choices

if t.TYPE_CHECKING:
    from temporalio.worker import Interceptor
    from temporalio.runtime import PrometheusConfig, TelemetryConfig
    from temporalio.runtime import OpenTelemetryMetricTemporality, OpenTelemetryConfig
    from temporalio.contrib.opentelemetry import TracingInterceptor

if load.TYPE_CHECKING:
    import opentelemetry
    from opentelemetry.exporter.otlp.proto.grpc import trace_exporter as otlp_trace_exporter
    import temporalio.runtime as temporal_runtime

else:
    
    opentelemetry = load.LazyLoad("opentelemetry", "opentelemetry-sdk", install_missing=True)
    otlp_trace_exporter = load.LazyLoad("opentelemetry.exporter.otlp.proto.grpc.trace_exporter", "opentelemetry-exporter-otlp", install_missing=True)
    temporal_runtime = load.LazyLoad("temporalio.runtime", "temporalio", install_missing=True)


prom_alias_prefixes = [
    'prometheus', 'telemetry', 'monitoring',
]
otel_alias_prefixes = [
    'otel', 'opentelemetry', 'tracing',
]

class TemporalPrometheusSettings(BaseSettings):
    """
    Temporal Settings for Prometheus
    """
    disabled: t.Optional[bool] = Field(
        None, 
        description = "If true, the Prometheus endpoint will not be exposed",
        validation_alias = create_alias_choices('disabled', *prom_alias_prefixes, key_pos = 'suffix')
    )
    # Allow for explicitly enabling Prometheus
    enabled: t.Optional[bool] = Field(
        None, 
        description = "If true, the Prometheus port will be exposed",
        validation_alias = create_alias_choices('enabled', *prom_alias_prefixes, key_pos = 'suffix')
    )

    address: t.Optional[str] = Field(
        None, 
        description = "The address to bind the Prometheus server to",
        validation_alias = create_alias_choices('address', 'prometheus', key_pos = 'suffix')
    )

    counters_total_suffix: t.Optional[bool] = Field(
        None, 
        description = "If true, the total counters will be exposed",
        validation_alias = create_alias_choices('counters_total_suffix', 'prometheus', key_pos = 'suffix')
    )

    unit_suffix: t.Optional[bool] = Field(
        None, 
        description = "If true, the unit suffix will be appended to the metrics",
        validation_alias = create_alias_choices('unit_suffix', 'prometheus', key_pos = 'suffix')
    )

    durations_as_seconds: t.Optional[bool] = Field(
        None, 
        description = "If true, durations will be exposed as seconds",
        validation_alias = create_alias_choices('durations_as_seconds', 'prometheus', key_pos = 'suffix')
    )

    class Config(BaseSettings.Config):
        env_prefix = "TEMPORAL_PROMETHEUS_"
    
    @model_validator(mode = 'after')
    def validate_prometheus(self):
        """
        Validates the Prometheus
        """
        if self.enabled:
            if not self.address: self.address = '0.0.0.0:9000'
        return self

    @eproperty
    def is_enabled(self) -> bool:
        """
        Returns whether the Prometheus is enabled
        """
        if self.enabled is not None: return self.enabled
        return self.address is not None if self.disabled is None else not self.disabled


    def update(self, data: t.Dict[str, t.Any]):
        """
        Updates the data
        """
        for key, value in data.items():

            if key in {'address', 'disabled'} and value is not None:
                _ = self._extra.pop('is_enabled', None)

            if hasattr(self, key): setattr(self, key, value)

    def get_config(self, **kwargs) -> 'PrometheusConfig':
        """
        Returns the Prometheus Config
        """
        kws = {
            'bind_address': self.address,
            'counters_total_suffix': self.counters_total_suffix,
            'unit_suffix': self.unit_suffix,
            'durations_as_seconds': self.durations_as_seconds,
        }
        if kwargs: kws.update(kwargs)
        kws = {k:v for k,v in kws.items() if v is not None}
        from temporalio.runtime import PrometheusConfig
        return PrometheusConfig(**kws)
    
    def get_interceptor(self, **kwargs) -> None:
        """
        Returns the Interceptors
        """
        return None

    def init_runtime(
        self, 
        **kwargs,
    ):
        """
        Initializes the Runtime for Prometheus
        """
        pass

class TemporalOTELSettings(BaseSettings):
    """
    Temporal Settings for OTEL
    """
    disabled: t.Optional[bool] = Field(
        None, 
        description = "If true, OpenTelemetry will not be enabled",
        validation_alias = create_alias_choices('disabled', *otel_alias_prefixes, key_pos = 'suffix')
    )

    url: t.Optional[str] = Field(
        None, 
        description = "The URL to use for the OTEL exporter",
        validation_alias = create_alias_choices('url', *otel_alias_prefixes, key_pos = 'suffix')
    )

    headers: t.Optional[Json] = Field(
        None, 
        description = "The headers to use for the OTEL exporter",
        validation_alias = create_alias_choices('headers', *otel_alias_prefixes, key_pos = 'suffix')
    )

    periodicity: t.Optional[str] = Field(
        'minutes=1', 
        description = "The metric periodicity to use for the OTEL exporter. Format should be comma-seperated <period>=<value>",
        validation_alias = create_alias_choices('periodicity', *otel_alias_prefixes, key_pos = 'suffix'),
        examples = [
            'minutes=1,seconds=30',
            'seconds=30',
        ]
    )

    temporality: t.Optional[str] = Field(
        None, 
        description = "The metric temporality to use for the OTEL exporter. Options are: CUMULATIVE, DELTA",
        validation_alias = create_alias_choices('temporality', *otel_alias_prefixes, key_pos = 'suffix'),
    )

    durations_as_seconds: t.Optional[bool] = Field(
        None, 
        description = "If true, durations will be exposed as seconds",
        validation_alias = create_alias_choices('durations_as_seconds',  *otel_alias_prefixes, key_pos = 'suffix')
    )

    service_name: t.Optional[str] = Field(
        None, 
        description = "The service name to use for the OTEL exporter",
        validation_alias = create_alias_choices('service_name', *otel_alias_prefixes, key_pos = 'suffix')
    )

    class Config(BaseSettings.Config):
        env_prefix = "TEMPORAL_OTEL_"
    
    @model_validator(mode = 'after')
    def validate_otel(self):
        """
        Validates the OTEL
        """
        if self.url:
            if not self.url.startswith('http'):
                if 'cluster.local' in self.url or 'localhost' in self.url:
                    self.url = f'http://{self.url}'
                else:
                    self.url = f'https://{self.url}'
        
        return self
    

    @eproperty
    def enabled(self) -> bool:
        """
        Returns whether the OTEL is enabled
        """
        return self.url is not None if self.disabled is None else not self.disabled
    
    @eproperty
    def insecure(self) -> bool:
        """
        Returns whether the OTEL is insecure
        """
        return self.url.startswith('http://') if self.url else False

    @eproperty
    def metric_periodicity(self) -> timedelta:
        """
        Returns the metric periodicity
        """
        kws = {period: int(value) for period, value in self.periodicity.split(',')}
        return timedelta(**kws)

    @eproperty
    def metric_temporality(self) -> t.Optional['OpenTelemetryMetricTemporality']:
        """
        Returns the metric temporality
        """
        if self.temporality is None: return None
        from temporalio.runtime import OpenTelemetryMetricTemporality
        return OpenTelemetryMetricTemporality[self.temporality.upper()]
    
    def update(self, data: t.Dict[str, t.Any]):
        """
        Updates the data
        """
        for key, value in data.items():
            if key == 'periodicity':
                if isinstance(value, timedelta):
                    self._extra['metric_periodicity'] = value
                    continue
                _ = self._extra.pop('metric_periodicity', None)
            
            if key == 'temporality':
                if hasattr(value, 'value'):
                    self._extra['metric_temporality'] = value.value
                    continue
                _ = self._extra.pop('metric_temporality', None)
            
            if key == 'headers':
                if isinstance(value, str): value = json.loads(value)
            
            if key in {'url', 'disabled'} and value is not None:
                _ = self._extra.pop('enabled', None)

            if hasattr(self, key): setattr(self, key, value)

    def get_config(self, **kwargs) -> 'OpenTelemetryConfig':
        """
        Returns the OTEL Config
        """
        kws = {
            'url': self.url,
            'headers': self.headers,
            'metric_periodicity': self.metric_periodicity,
            'metric_temporality': self.metric_temporality,
            'durations_as_seconds': self.durations_as_seconds,
        }
        if kwargs: kws.update(kwargs)
        kws = {k:v for k,v in kws.items() if v is not None}
        from temporalio.runtime import OpenTelemetryConfig
        return OpenTelemetryConfig(**kws)
    
    def get_interceptor(self, init: t.Optional[bool] = None, **kwargs) -> 'TracingInterceptor' | t.Type['TracingInterceptor']:
        """
        Returns the Interceptors
        """
        from temporalio.contrib.opentelemetry import TracingInterceptor
        return TracingInterceptor() if init else TracingInterceptor
    
    @eproperty
    def _has_init_runtime(self) -> bool:
        """
        Returns whether the Runtime has been initialized
        """
        return False

    def init_runtime(
        self, 
        service_name: t.Optional[str] = None,
        **kwargs,
    ):
        """
        Initializes the Runtime for OTEL
        """
        if not self.enabled: return
        service_name = service_name or self.service_name
        if not service_name: raise ValueError('Service Name must be provided')
        if self._has_init_runtime: return

        opentelemetry.__reload__()
        otlp_trace_exporter.__reload__()
        from opentelemetry import trace
        from opentelemetry.sdk.resources import SERVICE_NAME, Resource
        from opentelemetry.sdk.trace import TracerProvider
        from opentelemetry.sdk.trace.export import BatchSpanProcessor

        # Setup global tracer for workflow traces
        provider = TracerProvider(resource = Resource.create({SERVICE_NAME: service_name}))
        exporter = otlp_trace_exporter.OTLPSpanExporter(
            endpoint = self.url, 
            insecure = self.insecure,
            headers = self.headers or None,
        )
        provider.add_span_processor(BatchSpanProcessor(exporter))
        trace.set_tracer_provider(provider)
        self._has_init_runtime = True



class TemporalTelemetrySettings(BaseSettings):
    """
    Temporal Settings for Telemetry
    """

    @eproperty
    def prometheus(self) -> TemporalPrometheusSettings:
        """
        Returns the Prometheus Settings
        """
        return TemporalPrometheusSettings()
    
    @eproperty
    def otel(self) -> TemporalOTELSettings:
        """
        Returns the OTEL Settings
        """
        return TemporalOTELSettings()

    def update(self, data: t.Dict[str, t.Dict[str, t.Any]]):
        """
        Updates the data
        """
        for key, value in data.items():
            if key in {'prometheus', 'prom'}:
                self.prometheus.update(value)
                continue
            if key in {'otel', 'opentelemetry'}:
                self.otel.update(value)
                continue
    
    def get_telemetry_config(
        self, 
        provider: t.Optional[str] = None,
        init_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        config_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        init_interceptor: t.Optional[bool] = None,
        **kwargs
    ) -> t.Optional['TelemetryConfig'] | t.Optional[t.Union['Interceptor', t.Type['Interceptor']]]: 
        """
        Returns the Telemetry Runtime
        """
        if provider and provider.lower() in {'prometheus', 'prom'}:
            telm = self.prometheus
        elif provider and provider.lower() in {'otel', 'opentelemetry'}:
            telm = self.otel
        elif self.otel.enabled: telm = self.otel
        elif self.prometheus.is_enabled: telm = self.prometheus
        else: telm = None
        if telm is None: return None, None
        telm.init_runtime(**(init_kwargs or {}))
        return temporal_runtime.TelemetryConfig(
            metrics = telm.get_config(**(config_kwargs or {})),
            **kwargs,
        ), telm.get_interceptor(init = init_interceptor)


from __future__ import annotations

import json
import pathlib
import random
from lzl.api import aiohttpx
from lzl.types import BaseModel, Field, eproperty, model_validator, field_validator, validator, root_validator, PYDANTIC_VERSION
from lzl.proxied import proxied, ProxyObject
# from lzo.types import AppEnv


from lzl.api.openai.assets import load_preset_config, load_provider_prices
from lzl.api.openai.types.base import ModelCostItem
from lzl.api.openai.types.handlers import ModelContextHandler
from typing import Optional, Dict, Union, Any, List, Type, TYPE_CHECKING


# from lazyops.types import BaseSettings, validator, BaseModel, lazyproperty, Field, PYD_VERSION
# from lazyops.libs.proxyobj import ProxyObject
# from lazyops.libs.abcs.configs.types import AppEnv
# from async_openai.version import VERSION
# from async_openai.types.options import ApiType
# from async_openai.types.context import ModelCostItem, ModelContextHandler

# if PYD_VERSION == 2:
#     from pydantic import model_validator
# else:
#     from lazyops.types.models import root_validator

if TYPE_CHECKING:
    from lzl.api.openai.types.routes import BaseRoute, BaseResource


# preset_path = pathlib.Path(__file__).parent.joinpath('presets')


class ExternalProviderConfig(BaseModel):
    """
    External Provider - Configuration
    """
    api_base: str = Field(..., description="The base URL for the API")
    api_path: Optional[str] = Field(None, description="The path for the API")
    api_key_header: Optional[str] = Field(None, description="The header for the API Key")
    api_key_scheme: Optional[str] = Field(None, description="The scheme for the API Key")
    api_key: Optional[str] = Field(None, description="The API Key")
    api_keys: Optional[List[str]] = Field(None, description="The API Keys")

    max_retries: Optional[int] = Field(None, description="The maximum number of retries for the API")
    weight: Optional[float] = Field(None, description="The weight of the provider")

    custom_headers: Optional[Dict[str, str]] = Field(None, description="Custom Headers")
    proxy_url: Optional[str] = Field(None, description="The Proxy URL")
    proxy_headers: Optional[Dict[str, str]] = Field(None, description="Proxy Headers")
    hf_compatible: Optional[bool] = Field(None, description="Whether the provider is HuggingFace Compatible for Tokenization")
    supports_tokenization: Optional[bool] = Field(True, description="Whether the provider supports tokenization")

    # @validator("custom_headers", "proxy_headers", pre=True) 
    @field_validator("custom_headers", "proxy_headers", mode = 'before')
    def validate_headers(cls, value: Optional[Dict[str, Any]]) -> Optional[Dict[str, str]]:
        """
        Validates the Headers
        """
        if value is None: return {}
        value = {k: v for k,v in value.items() if v is not None}
        for k,v in value.items():
            value[k.strip()] = json.dumps(v) if isinstance(v, (list, dict)) else str(v).strip()
        return value
    
    # @validator("max_retries", pre=True)
    @field_validator("max_retries", mode = 'before')
    def validate_max_retries(cls, value: Optional[Union[str, int]]) -> Optional[int]:
        """
        Validates the Max Retries
        """
        if value is None: return None
        try:
            return int(value)
        except ValueError:
            return None
        
    # @validator("weight", pre=True)
    @field_validator("weight", mode = 'before')
    def validate_weight(cls, value: Optional[Union[str, float]]) -> Optional[float]:
        """
        Validates the Weight
        """
        if value is None: return None
        try: return float(value)
        except ValueError:
            return None

    @property
    def has_api_key(self) -> bool:
        """
        Returns whether the API Key is present
        """
        return self.api_key is not None
    
    @property
    def api_key_value(self) -> Optional[str]:
        """
        Returns the API Key Value
        """
        if not self.has_api_key: return None
        return f"{self.api_key_scheme} {self.api_key}" if self.api_key_scheme else self.api_key
    
    @property
    def has_api_keys(self) -> bool:
        """
        Returns whether the API Keys are present
        """
        return self.api_keys is not None
    
    @property
    def api_keys_value(self) -> Optional[List[str]]:
        """
        Returns the API Keys Value
        """
        if not self.has_api_keys: return None
        return [f"{self.api_key_scheme} {k}" if self.api_key_scheme else k for k in self.api_keys]

    @property
    def has_proxy(self) -> bool:
        """
        Returns whether the Proxy is present
        """
        return self.proxy_url is not None
    
    @property
    def api_url(self) -> str:
        """
        Returns the API URL
        """
        if not self.api_path: return self.api_base
        return aiohttpx.urljoin(self.api_base, self.api_path)


class ProviderModel(ModelCostItem):
    """
    A Provider Model
    """
    name: str = Field(..., description="The Model Name")
    tokenizer: Optional[str] = Field(None, description="The Tokenizer Name") 


class ProviderRoute(BaseModel):
    """
    A Provider Route
    """
    name: Optional[str] = Field(None, description="The Route Name")
    api_resource: Optional[str] = Field(None, description="The Route Path")
    root_name: Optional[str] = Field(None, description="The Root Name")
    # params: Optional[Dict[str, Type]] = Field(None, description="The Route Parameters and expected Types")

    if TYPE_CHECKING:
        route_class: Optional[Type['BaseRoute']] = Field(None, description="The Route Class")
        object_class: Optional[Type['BaseResource']] = Field(None, description="The Resource Class")
        response_class: Optional[Type['BaseResource']] = Field(None, description="The Response Class")
    else:
        route_class: Optional[Any] = Field(None, description="The Route Class")
        object_class: Optional[Any] = Field(None, description="The Resource Class")
        response_class: Optional[Any] = Field(None, description="The Response Class")
    

    # @validator("route_class", "object_class", "response_class", pre=True)
    @field_validator("route_class", "object_class", "response_class", mode = 'before')
    def validate_route_classes(cls, value: Union[str, Type]) -> Type:
        """
        Validates the Route Classes
        """
        if value is None: return None
        if isinstance(value, str):
            from lzl.load import lazy_import
            return lazy_import(value)
        return value


class ExternalProviderRoutes(BaseModel):
    """
    External Provider - Routes
    """
    completions: Optional[ProviderRoute] = Field(None, description="The Completion Route")
    chat: Optional[ProviderRoute] = Field(None, description="The Chat Route")
    embeddings: Optional[ProviderRoute] = Field(None, description="The Embedding Route")
    rerankings: Optional[ProviderRoute] = Field(None, description="The Reranking Route")
    # if TYPE_CHECKING:

    if PYDANTIC_VERSION == 2:
        @model_validator(mode = 'after')
        def validate_routes(self):
            """
            Validates the Routes
            """
            from lzl.api.openai.schemas import (
                ChatRoute, ChatResponse, ChatObject,
                CompletionRoute, CompletionResponse, CompletionObject,
                EmbeddingRoute, EmbeddingResponse, EmbeddingObject,
                RerankingRoute, RerankingResponse, RerankingObject,
                # TODO - Add the rest of the routes
            )
            # from async_openai.schemas import (
            #     ChatRoute, ChatResponse, ChatObject,
            #     CompletionRoute, CompletionResponse, CompletionObject,
            #     EmbeddingRoute, EmbeddingResponse, EmbeddingObject,
            #     # TODO - Add the rest of the routes
            # )
            if self.chat is None:
                self.chat = ProviderRoute(
                    name = "chat",
                    route_class = ChatRoute,
                    object_class = ChatObject,
                    response_class = ChatResponse,
                )
            if not self.chat.name:
                self.chat.name = "chat"
            if self.chat.route_class is None:
                self.chat.route_class = ChatRoute
            if self.chat.object_class is None:
                self.chat.object_class = ChatObject
            if self.chat.response_class is None:
                self.chat.response_class = ChatResponse

            if self.completions is None:
                self.completions = ProviderRoute(
                    name = "completions",
                    route_class = CompletionRoute,
                    object_class = CompletionObject,
                    response_class = CompletionResponse,
                )
            if not self.completions.name:
                self.completions.name = "completions"
            if self.completions.route_class is None:
                self.completions.route_class = CompletionRoute
            if self.completions.object_class is None:
                self.completions.object_class = CompletionObject
            if self.completions.response_class is None:
                self.completions.response_class = CompletionResponse

            if self.embeddings is None:
                self.embeddings = ProviderRoute(
                    name = "embeddings",
                    route_class = EmbeddingRoute,
                    object_class = EmbeddingObject,
                    response_class = EmbeddingResponse,
                )
            
            if not self.embeddings.name:
                self.embeddings.name = "embeddings"
            if self.embeddings.route_class is None:
                self.embeddings.route_class = EmbeddingRoute
            if self.embeddings.object_class is None:
                self.embeddings.object_class = EmbeddingObject
            if self.embeddings.response_class is None:
                self.embeddings.response_class = EmbeddingResponse
            
            if self.rerankings is None:
                self.rerankings = ProviderRoute(
                    name = "rerankings",
                    route_class = RerankingRoute,
                    object_class = RerankingObject,
                    response_class = RerankingResponse,
                )
            if not self.rerankings.name:
                self.rerankings.name = "rerankings"
            if self.rerankings.route_class is None:
                self.rerankings.route_class = RerankingRoute
            if self.rerankings.object_class is None:
                self.rerankings.object_class = RerankingObject
            if self.rerankings.response_class is None:
                self.rerankings.response_class = RerankingResponse
            
            return self

    else:
        @root_validator()
        def validate_routes(cls, values: Dict[str, Any]) -> Dict[str, Any]:
            """
            Validates the Routes
            """
            from lzl.api.openai.schemas import (
                ChatRoute, ChatResponse, ChatObject,
                CompletionRoute, CompletionResponse, CompletionObject,
                EmbeddingRoute, EmbeddingResponse, EmbeddingObject,
                RerankingRoute, RerankingResponse, RerankingObject,
                # TODO - Add the rest of the routes
            )
            # from async_openai.schemas import (
            #     ChatRoute, ChatResponse, ChatObject,
            #     CompletionRoute, CompletionResponse, CompletionObject,
            #     EmbeddingRoute, EmbeddingResponse, EmbeddingObject,
            #     # TODO - Add the rest of the routes
            # )
            if values.get('chat') is None:
                values['chat'] = ProviderRoute(
                    name = "chat",
                    route_class = ChatRoute,
                    object_class = ChatObject,
                    response_class = ChatResponse,
                )
            else:
                values['chat'] = ProviderRoute.parse_obj(values['chat'])
            if not values['chat'].name:
                values['chat'].name = "chat"
            if values['chat'].route_class is None:
                values['chat'].route_class = ChatRoute
            if values['chat'].object_class is None:
                values['chat'].object_class = ChatObject
            if values['chat'].response_class is None:
                values['chat'].response_class = ChatResponse
            
            if values.get('completions') is None:
                values['completions'] = ProviderRoute(
                    name = "completions",
                    route_class = CompletionRoute,
                    object_class = CompletionObject,
                    response_class = CompletionResponse,
                )
            else:
                values['completions'] = ProviderRoute.parse_obj(values['completions'])
            if not values['completions'].name:
                values['completions'].name = "completions"
            if values['completions'].route_class is None:
                values['completions'].route_class = CompletionRoute
            if values['completions'].object_class is None:
                values['completions'].object_class = CompletionObject
            if values['completions'].response_class is None:
                values['completions'].response_class = CompletionResponse
            
            if values.get('embeddings') is None:
                values['embeddings'] = ProviderRoute(
                    name = "embeddings",
                    route_class = EmbeddingRoute,
                    object_class = EmbeddingObject,
                    response_class = EmbeddingResponse,
                )
            else:
                values['embeddings'] = ProviderRoute.parse_obj(values['embeddings'])
            if not values['embeddings'].name:
                values['embeddings'].name = "embeddings"
            if values['embeddings'].route_class is None:
                values['embeddings'].route_class = EmbeddingRoute
            if values['embeddings'].object_class is None:
                values['embeddings'].object_class = EmbeddingObject
            if values['embeddings'].response_class is None:
                values['embeddings'].response_class = EmbeddingResponse

            if values.get('rerankings') is None:
                values['rerankings'] = ProviderRoute(
                    name = "rerankings",
                    route_class = RerankingRoute,
                    object_class = RerankingObject,
                    response_class = RerankingResponse,
                )
            else:
                values['rerankings'] = ProviderRoute.parse_obj(values['rerankings'])
            if not values['rerankings'].name:
                values['rerankings'].name = "rerankings"
            if values['rerankings'].route_class is None:
                values['rerankings'].route_class = RerankingRoute
            if values['rerankings'].object_class is None:
                values['rerankings'].object_class = EmbeddingObject
            if values['rerankings'].response_class is None:
                values['rerankings'].response_class = RerankingResponse
            return values
        
    @property
    def api_route_classes(self) -> Dict[str, Type['BaseRoute']]:
        """
        Returns the Route Classes
        """
        return {
            "chat": self.chat.route_class,
            "completions": self.completions.route_class,
            "embeddings": self.embeddings.route_class,
            "rerankings": self.rerankings.route_class,
        }
    
            

class ExternalProviderSettings(BaseModel):
    """
    External Provider Configuration
    """
    name: str = Field(..., description="The Provider Name")
    config: ExternalProviderConfig = Field(default_factory = ExternalProviderConfig, description="The Provider Configuration")
    models: List[ProviderModel] = Field(default_factory=list, description="The Provider Models")
    routes: Optional[ExternalProviderRoutes] = Field(default_factory=ExternalProviderRoutes, description="The Provider Routes")

    @classmethod
    def from_preset(
        cls, 
        name: Optional[str] = None,
        path: Optional[Union[str, pathlib.Path]] = None,
        **overrides
    ) -> 'ExternalProviderSettings':
        """
        Loads the Provider Settings from a Preset
        """
        data = load_preset_config(name, path, **overrides)
        provider_settings = cls.model_validate(data)
        ModelContextHandler.add_provider(provider_settings)
        return provider_settings
    
    @property
    def model_list(self) -> List[str]:
        """
        Returns the Model List (excludes the provider name)
        """
        names = [m.name for m in self.models]
        # Add aliases
        for m in self.models:
            if m.aliases:
                names.extend(m.aliases)
        return names
    

    # if PYD_VERSION == 2:
    #     @model_validator(mode = 'after')
    #     def validate_provider_settings(self):
    #         """
    #         Validate the provider settings
    #         """
    #         if self.config.max_retries is not None:
    #             # from lazyops.utils.logs import logger
    #             # logger.info(f'Configuring max retries: {self.config.max_retries}')
    #             self.routes.configure_retry(max_retries = self.config.max_retries)
    #         return self

